"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit visibility graphics settings. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects import result as res

from duHast.Utilities.files_json import read_json_data_from_file

from duHast.Revit.Views.filters import apply_filter_to_view
from duHast.Revit.Views.Objects.view_graphics_settings import ViewGraphicsSettings
from duHast.Revit.Views.Reporting.views_data_report import PROP_VIEW_DATA
from duHast.Revit.Views.templates import get_view_templates
from duHast.Revit.Views.Utility.convert_revit_override_to_data import get_view_settings
from duHast.Revit.Views.Utility.convert_data_to_override_storage import (
    convert_to_category_override_storage_objects,
    convert_to_filter_override_storage_objects,
)
from duHast.Revit.Views.visibility_graphics_filters import apply_filter_override_to_view
from duHast.Revit.Views.visibility_graphics_categories import (
    apply_graphic_override_to_view,
)

from duHast.Revit.LinePattern.line_patterns import build_patterns_dictionary_by_name
from duHast.Revit.LinePattern.fill_patterns import pattern_ids_by_name
from duHast.Revit.Common.Objects.Data.pattern_settings_base import PatternSettingBase

from Autodesk.Revit.DB import Element, LinePatternElement


def import_graphic_overrides(file_path, call_back):
    """
    Reads JSON data from a file and converts it into a dictionary of ViewGraphicsSettings objects.

    Args:
        file_path (str): The fully qualified file path of the JSON file.
        call_back (function): A callback function to update the progress of the data import.

    Returns:
        dict: A dictionary of ViewGraphicsSettings objects, where the view name is the key and the ViewGraphicsSettings object is the value.
    """
    return_value = res.Result()
    # print("ini: {}".format(return_value))
    view_overrides_data = {}
    read_result = read_json_data_from_file(file_path)
    if read_result.status is False:
        return_value.update_sep(False, read_result.message)
        return return_value
    json_data = read_result.result[0]
    
    counter = 1
    if PROP_VIEW_DATA in json_data:
        for json_override in json_data[PROP_VIEW_DATA]:
            view_override = ViewGraphicsSettings(j=json_override)
            view_overrides_data[view_override.view_name] = view_override
            return_value.append_message(
                "reading template: {} :: {} of {}".format(
                    view_override.view_name, counter, len(json_data[PROP_VIEW_DATA])
                )
            )
            counter = counter + 1
            # call back update
            if call_back:
                call_back(counter, len(json_data[PROP_VIEW_DATA]))
    else:
        return_value.update_sep(
            False, "File {} contains no view data.".format(file_path)
        )
    return_value.result.append(view_overrides_data)
    return return_value


def _check_pattern_in_model(pattern_from_model, pattern_data):
    """
    Check whether the line or fill patterns used in the given pattern data exist in the Revit model.

    Args:
        pattern_from_model (dict): A dictionary containing the line or fill patterns in the Revit model.
        pattern_data (dict): A dictionary containing the line or fill patterns to be checked.

    Returns:
        Result: A Result object containing the results of the pattern check.
    """

    return_value = res.Result()
    no_match = []

    for key, value in pattern_data.items():
        # ignore any pattern with -1 as id...its the default for no pattern assigned
        if value.id != -1:
            if key not in pattern_from_model:
                return_value.update_sep(
                    False, "{}......does not exist in model.".format(key)
                )
                no_match.append(key)
            else:
                return_value.update_sep(True, "{}......exists in model.".format(key))
        else:
            return_value.update_sep(True, "{}......ignored -1".format(key))

    if not return_value.status:
        return_value.result.append(no_match)

    return return_value


def check_line_patterns_are_in_model(doc, line_pattern_data):
    """
    Check whether the line patterns used in the `line_pattern_data` exist in the Revit model.

    Args:
        doc (Revit Document): The Revit document object.
        line_pattern_data (dictionary): A dictionary containing line pattern data.

    Returns:
        Result: A Result object containing the results of the line pattern check.
    """
    return_value = res.Result()
    all_line_pattern_in_model = build_patterns_dictionary_by_name(doc=doc)
    # append solid line pattern as it is  a special case inn the api
    all_line_pattern_in_model[PatternSettingBase.SOLID_PATTERN] = [
        LinePatternElement.GetSolidPatternId()
    ]
    # check line pattern in model against line pattern in data
    return_value.update(
        _check_pattern_in_model(
            pattern_from_model=all_line_pattern_in_model, pattern_data=line_pattern_data
        )
    )
    return return_value


def check_fill_patterns_are_in_model(doc, fill_pattern_data):
    """
    Check whether all fill patterns used in the fill_pattern_data exist in the Revit model.

    Args:
        doc (Revit Document): The Revit document object.
        fill_pattern_data (dict): A dictionary containing fill pattern data.

    Returns:
        Result: A Result object containing the results of the fill pattern check.
    """
    return_value = res.Result()
    all_fill_pattern_in_model = pattern_ids_by_name(doc=doc)
    # check fill pattern in model against fill pattern data
    return_value.update(
        _check_pattern_in_model(
            pattern_from_model=all_fill_pattern_in_model, pattern_data=fill_pattern_data
        )
    )
    return return_value


def check_all_line_and_fill_pattern_in_model(doc, view_overrides_data):
    """
    Check whether all line patterns and fill patterns used in the view_overrides_data exist in the model.

    Args:
        doc (Revit Document): The Revit document object.
        view_overrides_data (dict): A dictionary containing view overrides data.

    Returns:
        Result: A Result object containing the results of the line pattern and fill pattern checks.
    """
    return_value = res.Result()
    for key, view_override in view_overrides_data.items():
        return_value.append_message(
            "Checking fill patterns from template: {} from file against patterns from model.".format(
                view_override.view_name
            )
        )
        # check whether all line patterns and fill pattern exist in model
        return_value.update(
            check_line_patterns_are_in_model(
                doc=doc,
                line_pattern_data=view_override.get_all_used_line_patterns(),
            )
        )
        return_value.append_message(
            "Checking line patterns from template: {} from file against fill patterns from model.".format(
                view_override.view_name
            )
        )
        return_value.update(
            check_fill_patterns_are_in_model(
                doc=doc,
                fill_pattern_data=view_override.get_all_used_fill_patterns(),
            )
        )
    return return_value


def get_matching_templates(doc, view_overrides_data):
    """
    Returns a dictionary containing the view templates that match the view names specified in the view_overrides_data dictionary.

    Args:
        doc (Autodesk.Revit.DB.Document): The current Revit model document.
        view_overrides_data (dict): A dictionary containing view override data. The keys are view names and the values are objects of type ViewOverrideData.

    Returns:
        dict: A dictionary containing the view templates that match the view names specified in the view_overrides_data dictionary. The keys are view names and the values are tuples containing the view template and the corresponding view override data.
    """
    matching_templates = {}
    view_templates_in_model = get_view_templates(doc)
    for key, view_override in view_overrides_data.items():
        for template in view_templates_in_model:
            if Element.Name.GetValue(template) == view_override.view_name:
                matching_templates[view_override.view_name] = (template, view_override)
                break

    return matching_templates


def _add_filters_to_view(doc, view, filter_data):
    """
    Adds filters to a view in Autodesk Revit.

    Args:
        doc (Autodesk.Revit.DB.Document): The document object representing the Revit model.
        view (Autodesk.Revit.DB.View): The view object to which the filters should be added.
        filter_data (list): A list of filter objects that should be added to the view.

    Returns:
        duHast.Utilities.Objects.result.Result: The result object containing the outcome of adding the filters to the view.
        The `status` attribute of the result object indicates whether the filters were added successfully or not.
        The `message` attribute of the result object contains any additional messages related to the application of the filters.
        The `result` attribute of the result object is a list of filters that were successfully added to the view.
    """
    return_value = res.Result()
    filters_exist_in_view = []
    for filter in filter_data:
        status_filter = apply_filter_to_view(doc=doc, view=view, filter=filter.filter)
        return_value.update(status_filter)
        if status_filter.status:
            filters_exist_in_view.append(filter)
    return_value.result.append(filters_exist_in_view)
    return return_value


def apply_override_to_view(doc, view, view_override):
    """
    Apply graphic and filter overrides to a Revit view.

    Args:
        doc (Revit Document): The current Revit document.
        view (Revit View): The view to apply the overrides to.
        view_override (ViewOverride): An object containing the desired graphic and filter overrides.

    Returns:
        Result: A result object containing the status of the override application, any error messages, and additional information.
    """
    return_value = res.Result()
    return_value.append_message("Modifying view: {}".format(view.Name))
    target_view_data = get_view_settings(doc=doc, view=view)
    # find overrides requiring updates
    category_overrides_to_update = view_override.get_differing_category_overrides(
        other_view_graphic_settings=target_view_data
    )
    # if import has different filters to target update existing filter overrides only
    filter_overrides_to_update = view_override.get_differing_filter_overrides(
        other_view_graphic_settings=target_view_data
    )
    # update target template
    # category overrides
    if len(category_overrides_to_update) > 0:
        return_value.append_message(
            "Found category overrides to update: {}".format(
                len(category_overrides_to_update)
            )
        )
        # convert to override storage objects for ease of applying to view
        category_overrides_to_apply = convert_to_category_override_storage_objects(
            doc=doc, category_data_objects=category_overrides_to_update
        )
        # apply overrides to view
        result_apply_cat_overrides = apply_graphic_override_to_view(
            doc=doc,
            view=view,
            category_storage_instances=category_overrides_to_apply,
        )
        return_value.update(result_apply_cat_overrides)
    else:
        return_value.append_message("No category overrides needed updating")

    # filter overrides
    if len(filter_overrides_to_update) > 0:
        return_value.append_message(
            "Found filter overrides to update: {}".format(
                len(filter_overrides_to_update)
            )
        )

        # convert filter override to filter override storage object
        filter_overrides_to_apply = convert_to_filter_override_storage_objects(
            doc=doc, filter_data_objects=filter_overrides_to_update
        )
        # check if filter is applied to template if not add them in
        # this is not required since the filters to apply check only returns filters which are
        # applied to the source (from disk) and destination (in model) views
        filter_overrides_to_apply_status = _add_filters_to_view(
            doc=doc,
            view=view,
            filter_data=filter_overrides_to_apply,
        )
        return_value.update_sep(
            filter_overrides_to_apply_status.status,
            filter_overrides_to_apply_status.message,
        )
        if filter_overrides_to_apply_status.status:
            # apply filter override
            result_apply_filter_overrides = apply_filter_override_to_view(
                doc=doc,
                view=view,
                filter_storage_instances=filter_overrides_to_apply_status.result[0],
            )
            return_value.update(result_apply_filter_overrides)
    else:
        return_value.append_message("No filter overrides needed updating")
    return return_value


def apply_overrides_to_views(doc, view_data):
    """
    Applies graphic and filter overrides to multiple views in a Revit document.

    Args:
        doc (Revit Document): The current Revit document.
        view_data (dict): A dictionary where the keys are the names of the views and the values are lists containing the Revit view object and the view override object.

    Returns:
        Result: A result object containing the status of the override application, any error messages, and additional information.
    """

    return_value = res.Result()
    for key, value in view_data.items():
        status_apply = apply_override_to_view(doc, value[0], value[1])
        return_value.update(status_apply)
    return return_value


def apply_overrides_from_file(doc, file_path):
    """
    Applies graphic and filter overrides to multiple views in a Revit document based on data imported from a JSON file.

    :param doc: The current Revit document.
    :type doc: Revit Document
    :param file_path: The fully qualified file path of the JSON file containing the view override data.
    :type file_path: str

    :returns: A result object containing the status of the override application, any error messages, and additional information.
    :rtype: Result

    :example:

    Example Usage:
    >>> result = apply_overrides_from_file(doc, file_path)
    >>> if result.status:
    ...     print("Overrides applied successfully.")
    >>> else:
    ...     print("Failed to apply overrides:", result.message)
    """

    return_value = res.Result()
    try:
        # load data from file
        view_overrides_data_status = import_graphic_overrides(
            file_path=file_path, call_back=None
        )
        return_value.update(view_overrides_data_status)
        # check if any data was retrieved
        if view_overrides_data_status.status:
            view_overrides_data = view_overrides_data_status.result[0]
            # check whether all line patterns and fill pattern exist in model
            pattern_status = check_all_line_and_fill_pattern_in_model(
                doc=doc, view_overrides_data=view_overrides_data
            )
            return_value.update(pattern_status)
            if pattern_status.status:
                # find matching templates
                matching_templates = get_matching_templates(
                    doc=doc, view_overrides_data=view_overrides_data
                )
                if len(matching_templates) > 0:
                    apply_overrides = apply_overrides_to_views(
                        doc=doc, view_data=matching_templates
                    )
                    return_value.update(apply_overrides)
                else:
                    return_value.append_message(
                        "No matching view templates found in file."
                    )
            else:
                # Flatten the list of lists into a single list
                flattened_list = [
                    item for sublist in pattern_status.result for item in sublist
                ]
                # build list of missing patterns and raise error
                raise ValueError(
                    "Aborted: The following patterns are missing from the file: {}".format(
                        ",".join(flattened_list)
                    )
                )
        else:
            raise ValueError(view_overrides_data_status.message)

    except Exception as e:
        return_value.update_sep(
            False, "Failed to apply overrides with exception: {}".format(e)
        )
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Utility functions converting collector objects to blueprint objects.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects.result import Result
from duHast.Data.Objects.Collectors.data_sheet import DataSheet

def sheets_group_by_instance_property(sheet_data, property_name):
    """
    Groups sheets by a given property

    :param sheet_data: A list od DataSheet instances
    :type sheet_data: [:class: `.DataSheet`]
    :param property_name: The name of the sheet instance property by its value to group the sheets by
    :type property_name: str
    :raises ValueError: property_name needs to be of type str
    :raises ValueError: sheet_data needs to be of type list
    :raises ValueError: sheet got multiple properties of the given name
    :raises ValueError: sheet no properties of the given name
    :raises ValueError: sheet_data list contained a none DataSheet object

    :return:
            Result class instance.
                - .status True if sheets where grouped successfully, otherwise False.
                - .message will be empty.
                - . result (a list with the grouped sheets as a dictionary)

                on exception:

                - .result Will be False
                - .message will contain exception message.
                - . result (empty list)
        :rtype: :class:`.Result`
    """

    return_value = Result()
    
    matches = {}

    try:

        # do some type checking
        if(isinstance(property_name, str)==False):
            raise ValueError("property_name needs to be a string, got {} instead.".format(type(property_name)))
        if(isinstance(sheet_data, list)==False):
            raise ValueError("sheet_data needs to be a list, got {} instead.".format(type(sheet_data)))
        
        # loop over sheets and attempt to group
        for sheet in sheet_data:
            if(isinstance(sheet, DataSheet)):
                # Finding objects with the specific property value using list comprehension
                result = [prop for prop in sheet.instance_properties.properties if prop.name == property_name]
                if(len(result)==1):
                    value = result[0].value 
                    if value in matches:
                        matches[value].append(sheet)
                    else:
                        matches[value] = [sheet]
                elif(len(result)>1):
                    # this really should never happen
                    raise ValueError ("Sheet got multiple instance properties of name: {}".format(property_name))
                else:
                    raise ValueError("Sheet has no property of name: {}".format(property_name))
            else:
                raise ValueError("Data list contained non sheet element!")
    except Exception as e:
        return_value.update_sep(False, "failed to group sheets by property: {} with exception: {}".format(property_name, e))
    
    # rteturn groupings to user
    return_value.result.append(matches)
    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family line pattern purge unused utilities.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This will delete all line patterns which are not used by any element in the family or nested families.

- requires a revit line pattern processor object

"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

# class used for stats reporting
from duHast.Utilities.Objects import result as res
from duHast.Revit.Common import delete as rDel
from duHast.Revit.Family.Data.Objects import ifamily_data as IFamData
from duHast.Revit.LinePattern.Data.Objects import line_pattern_data as rLinePatData

import Autodesk.Revit.DB as rdb


def purge_unused(doc, processor):
    """
    This will delete all line patterns that are not used by any element in the family or nested families.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param processor: A LinePatternProcessor object containing all line pattern information of the family document and any nested families.
    :type processor: :class:`.LinePatternProcessor`

    :return: Result class instance.
        - `result.status` (bool): True if all unused line patterns were deleted successfully or none needed to be deleted. Otherwise, False.
        - `result.message` (str): Property updated in the format: "Found unused line pattern: line pattern Name [subcategory Id]"

    On exception:
        - `result.status` (bool) will be False.
        - `result.message` (str) will contain the exception message.

    :rtype: :class:`.Result`
    """

    # from processor instance get all root line pattern entries where usage counter == 0.
    # delete those line patterns by id

    return_value = res.Result()

    ids_to_delete = []
    # get categories found in root processor data only
    root_fam_data = processor._find_root_family_data()
    # get all root line pattern entries where usage counter == 0.
    for root_fam in root_fam_data:
        if root_fam[IFamData.USAGE_COUNTER] == 0:
            return_value.append_message(
                "Found unused line patterns: {} [{}]".format(
                    root_fam[rLinePatData.PATTERN_NAME],
                    root_fam[rLinePatData.PATTERN_ID],
                )
            )
            ids_to_delete.append(rdb.ElementId(root_fam[rLinePatData.PATTERN_ID]))
    # delete any subcategories found
    if len(ids_to_delete) > 0:
        result_delete = rDel.delete_by_element_ids(
            doc, ids_to_delete, "Deleting unused line patterns.", "Line patterns"
        )
        return_value.update(result_delete)
    else:
        return_value.update_sep(
            True, "No unused line patterns found. Nothing was deleted."
        )
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Reload using advanced tools collection.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module provides utility functions to read and write reload task lists for the Revit Batch Processor.

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#
import os
from collections import namedtuple

from duHast.Utilities import (
    files_get as fileGet,
    files_io as fileIO,
)
from duHast.Utilities.files_csv import write_report_data_as_csv, read_csv_file

# tuples containing base family data and changed family data read from files
changed_family = namedtuple("changed_family", "name category filePath")
# baseFamily = namedtuple('baseFamily', 'name category rootPath filePath')

# row structure of family change data file
CHANGE_LIST_INDEX_FAMILY_NAME = 0
CHANGE_LIST_INDEX_FAMILY_FILE_PATH = 1
CHANGE_LIST_INDEX_CATEGORY = 2

TASK_COUNTER_FILE_PREFIX = "TaskOutput"


def write_reload_list_to_file(reload_families, directory_path, counter=0):
    """
    Writes task list file to disk. File contains single column of fully qualified file path.

    :param reload_families: List of tuples representing families requiring their nested families to be re-loaded.
    :type reload_families: [baseFamily]
    :param directory_path: Fully qualified directory path to which the task files will be written.
    :type directory_path: str
    :param counter: Task file name suffix, defaults to 0
    :type counter: int, optional
    :return: True if file was written successfully, otherwise False.
    :rtype: bool
    """

    # write out file list without header
    header = []
    # data to be written to file
    overall_data = []
    file_name = os.path.join(
        directory_path, "{} {}.txt".format(TASK_COUNTER_FILE_PREFIX, counter)
    )
    # loop over families to get file path
    for r in reload_families:
        # row data
        data = []
        data.append(r.filePath)
        overall_data.append(data)
    try:
        # write data
        write_result = write_report_data_as_csv(file_name, header, overall_data, writeType="w")
        return write_result.status
    except Exception:
        return False


def delete_old_task_lists(directory_path):
    """
    Deletes all overall task files in given directory.

    :param directory_path: Fully qualified directory path containing the task files to be deleted.
    :type directory_path: str
    :return: True if all files got deleted successfully, otherwise False.
    :rtype: bool
    """

    flag = True
    # find all files in folder starting with and delete them
    files = fileGet.get_files(directory_path, ".txt")
    if len(files) > 0:
        for f in files:
            if fileIO.get_file_name_without_ext(f).startswith(TASK_COUNTER_FILE_PREFIX):
                flag = flag & fileIO.file_delete(f)
    return flag


def write_out_empty_task_list(directory_path, counter=0):
    """
    Writes out an empty task list in case nothing is to be reloaded.

    :param directory_path: Fully qualified directory path to which the task files will be written.
    :type directory_path: str
    :param counter: Task file name suffix, defaults to 0
    :type counter: int, optional
    :return: True if file was written successfully, otherwise False.
    :rtype: bool
    """

    file_name = os.path.join(
        directory_path, "{} {}.txt".format(TASK_COUNTER_FILE_PREFIX, counter)
    )
    # write out file list without header
    header = []
    # write out empty data
    overall_data = []
    try:
        # write data
        write_result = write_report_data_as_csv(file_name, header, overall_data, writeType="w")
        return write_result.status
    except Exception:
        return False


def _remove_rfa_from_file_name(family_name):
    """
    Removes any .rfa file extensions from the family name. (not sure why these are sometimes present)

    :param family_name: the family name
    :type family_name: str
    :return: the family name with out .rfa (if present in the first place.)
    :rtype: str
    """

    if family_name.lower().endswith(".rfa"):
        family_name = family_name[: -len(".rfa")]
    return family_name


def read_change_list(file_path):
    """
    Reads list of changed families from file into named tuples.

    :param file_path: Fully qualified file path to change list  file.
    :type file_path: str
    :raises Exception: "Changed families list files does not exist."
    :raises Exception: "Empty families list file!"
    :return: list of named tuples
    :rtype: [changed_family]
    """

    rows = []
    if fileIO.file_exist(file_path):
        # attempt to read file
        rows_result = read_csv_file(file_path)
        if rows_result.status is False:
            raise Exception("Error reading changed families list file.: {}".format(rows_result.message))
        rows = rows_result.result
    else:
        raise Exception("Changed families list files does not exist.")
    if len(rows) > 0:
        pass
    else:
        raise Exception("Empty families list file!")

    return_value = []
    # skip header row
    for i in range(1, len(rows)):
        # TODO: do i need any .rfa from end of family name?
        fam_name = _remove_rfa_from_file_name(rows[i][CHANGE_LIST_INDEX_FAMILY_NAME])
        data = changed_family(
            fam_name,
            rows[i][CHANGE_LIST_INDEX_CATEGORY],
            rows[i][CHANGE_LIST_INDEX_FAMILY_FILE_PATH],
        )
        return_value.append(data)
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit view filters to storage conversion. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
clr.AddReference('System')
from System.Collections.Generic import List

from duHast.Data.Utils.data_to_file import build_json_for_file
from duHast.Utilities.files_io import get_file_name_without_ext
from duHast.Utilities.Objects.result import Result
from duHast.Utilities.files_json import write_json_to_file


from duHast.Revit.Views.filters import get_all_filters
from duHast.Revit.Views.Objects.Data.view_filter import ViewFilter
from duHast.Revit.Views.Objects.Data.view_filter_rule import ViewFilterRule
from duHast.Revit.Views.Objects.Data.view_filter_logic_container import ViewFilterLogicContainer

from Autodesk.Revit.DB import Element, LogicalAndFilter, LogicalOrFilter, ElementFilter, ElementParameterFilter,  FilterNumericValueRule, FilterInverseRule, FilterStringRule


DEBUG = False


# Note:
# Inverse rules are essentially wrappers around standard rules inversing that outcome


def analyze_rule(doc, rule,  is_inversed, nesting_level,):
    

    view_filter_rule = ViewFilterRule()

    # can be numeric or a string rule
    # get the parameter to be checked
    if DEBUG:
        print ("{} rule parameter id: {}".format("..." * nesting_level, rule.GetRuleParameter().IntegerValue))
    view_filter_rule.parameter_id = rule.GetRuleParameter().IntegerValue

    # get the evaluation type (ends with, starts with, equals, greater than, etc)
    if DEBUG:
        print ("{} rule evaluator: {}".format("..." * nesting_level, type(rule.GetEvaluator())))
    view_filter_rule.evaluation_type = rule.GetEvaluator().GetType().Name

    # get the rule value
    if isinstance(rule, FilterNumericValueRule):
        if DEBUG:
            print ("{} rule value: {}".format("..." * nesting_level, rule.RuleValue))
        view_filter_rule.rule_value=rule.RuleValue
    elif isinstance(rule, FilterStringRule):
        if DEBUG:
            print ("{} rule value: {}".format("..." * nesting_level, rule.RuleString))
        view_filter_rule.rule_value=rule.RuleString
    
    # is the rule inversed
    if DEBUG:
        print ("{} is inversed: {}".format("..." * nesting_level, is_inversed))
    
    view_filter_rule.is_inversed = is_inversed

    if DEBUG:
        print ("{} rule complete: {}".format("..." * nesting_level, view_filter_rule))
    return view_filter_rule

    



def analyze_element_parameter_filter(doc, element_parameter_filter, nesting_level):

    rules = element_parameter_filter.GetRules()
    if DEBUG:
        print ("{} rules: {}".format("..." * nesting_level, rules.Count))
    
    # set up a list to hold the rules
    rules_analysed = []

    # go over all rules
    for rule in rules:
        if DEBUG:
            print ("{} rule type: {}".format("..." * nesting_level, type(rule)))
       
        # check if the rule is an inverse rule
        if isinstance(rule, FilterInverseRule):
            if DEBUG:
                print ("{} is inverse rule...unwrap".format("..." * nesting_level))
            rule_nested = rule.GetInnerRule()
            rule_analysed = analyze_rule(doc, rule_nested, True, nesting_level+1)
            
            # add generated rule to list
            rules_analysed.append(rule_analysed)
        else:
            rule_analysed = analyze_rule(doc, rule, False, nesting_level+1)
            
            # add generated rule to list
            rules_analysed.append(rule_analysed)
    
    return rules_analysed
       


def analyze_logical_filter(doc, logical_filter, nesting_level=0):
    
    logical_container = ViewFilterLogicContainer()

    # get the filters in the logical filter
    # should always be a list of element parameter filters or nested logical filters
    filters = logical_filter.GetFilters()

    if DEBUG:
        print ("{} Logical filter contains the following filters:".format("..." * nesting_level))
        print ("{} {} ".format("..." * nesting_level, type(filters)))

    # set up a comparing type
    filter_list_type = List[ElementFilter]

    # should always be a list of element parameter filters or nested logical filters
    if isinstance(filters, filter_list_type):
        for filter in filters:
            # check the type of filter
            if isinstance(filter, ElementParameterFilter):
                if DEBUG:
                    print ( "{} is element parameter filter".format("..." * nesting_level))
                rules = analyze_element_parameter_filter(doc, filter, nesting_level + 1)
                
                # check what came back
                if len(rules) > 0:
                    if DEBUG:
                        print("{} adding {} rules to logical container".format("..." * nesting_level, len(rules)))
                    # add rules to the logical container
                    logical_container.view_filter_rules = logical_container.view_filter_rules + rules

            elif isinstance(filter, LogicalAndFilter) or isinstance(filter_elements, LogicalOrFilter):
                if DEBUG:
                    print ( "{} is logical and filter...recursive call".format("..." * nesting_level))
                nested_container = analyze_logical_filter(doc,filter, nesting_level + 1)
                
                # check what came back
                if  nested_container:
                    # add nested container to the logical container
                    logical_container.logic_containers.append(nested_container)
                    
                
    else:
        if DEBUG:
            # not sure what this...
            print("{} Currently not supported: {}".format("..." * nesting_level, type(filters)))
        return none
    
    if DEBUG:
        print ("{} logical filter complete: {}".format("..." * nesting_level, logical_container))
    return logical_container




def analyze_filters(doc, filters, forms):


    analysed_filters = []

    max_value = len(filters.ToElements())
    counter = 1

    # set up a pyrevit progress bar
    with forms.ProgressBar(
        title="Exporting view filters: {value} of {max_value}", cancellable=True
    ) as pb:

        # loop over view filters in the model
        for filter in filters:
            
            # update progress bar
            pb.update_progress(counter, max_value)

            view_filter = ViewFilter()
            
            if DEBUG:
                # get the filter name
                print("filter name: {}".format(Element.Name.GetValue(filter)))
            view_filter.name = Element.Name.GetValue(filter)

            # getting the revit category ids the filter is applied to
            filter_revit_category_ids = filter.GetCategories()
            for id in filter_revit_category_ids:
                if DEBUG:
                    print("...Filter Id [{}]".format(id.IntegerValue))
                view_filter.category_ids.append(id.IntegerValue)
            
            # getting the filter elements
            filter_elements = filter.GetElementFilter()

            # check the type of filter, should be a logical element filter (top level)
            if isinstance(filter_elements, LogicalAndFilter) or isinstance(filter_elements, LogicalOrFilter):
                if DEBUG:
                    print ( "...is logical filter")
                container_host = analyze_logical_filter(doc, filter_elements,1)
                if container_host:
                    view_filter.logic_container = container_host
                    if DEBUG:
                        print("...updated container host to view filter")
                    analysed_filters.append(view_filter)
            else:
                if DEBUG:
                    # not sure what this...
                    print("    Currently not supported: {}".format(type(filter_elements)))

            if pb.cancelled:
                return None
            
            # update progress
            counter = counter + 1

    return analysed_filters



# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#



from duHast.Utilities.Objects.result import Result



from duHast.Revit.Family.LibraryCleanUp.Utility.defaults import GROUPING_CODE_PARAMETER_NAME, CATEGORY_FILE_NAME_PREFIX_MAPPER
from duHast.Revit.Family.LibraryCleanUp.Utility.grouping_code import clean_code, convert_to_file_name_code, load_group_code_description
from duHast.Revit.Family.LibraryCleanUp.Utility.family_file_name import build_family_name_from_descriptor


from duHast.Revit.Family.Data.Objects.family_type_data_storage_manager import FamilyTypeDataStorageManager
from duHast.Revit.Family.Data.Objects.family_type_data_storage import FamilyTypeDataStorage
from duHast.Revit.Family.Data.Objects.family_directive_copy import FamilyDirectiveCopy
from duHast.Revit.Family.Data.Objects.family_directive_swap_instances_of_type import FamilyDirectiveSwap

from duHast.Utilities.files_io import file_exist



def get_code_to_use(full_code, group_code, code_to_descriptor_map):

    # get the description for the group code
    code_description = code_to_descriptor_map.get(full_code, None)

    # this should not happen....
    if code_description is None:
        # try to get the group code description from the group code
        code_description = code_to_descriptor_map.get(group_code, None)
        if code_description is None:
            # if the group code is not in the mapping, raise an error
            raise ValueError("Grouping code '{}' and full code: {} not found in code description mapping.".format(group_code, full_code))
    
    # use the first part of the description as the code to use for the file name
    code_to_use  = code_description[0]
    return code_to_use, code_description

def get_unique_group_codes_from_family(family_storage_data):
    """
    Get unique group codes from the family storage data.

    will return a dictioanry where key is the full code and value is the group code without any sub codes.
    ITSE-002.01 will return {'ITSE-002.01': 'ITSE-002'}
    ITSE-001 will return {'ITSE-001': 'ITSE-001'}

    :param family_storage_data: The family storage data from which to get the unique group codes.
    :type family_storage_data: :class:`.FamilyTypeDataStorageManager`
    :return: A dictionary of unique  codes and their cleaned versions.
    :rtype: {str:str}
    """

    unique_group_codes = {}
   
    if (isinstance(family_storage_data, FamilyTypeDataStorageManager)==False):
        raise TypeError("family_storage_data must be an instance of FamilyTypeDataStorageManager. Got instead: {}".format(type(family_storage_data)))
    
    if (family_storage_data.family_has_types == False):
        # return an empty list if the family has no types
        return unique_group_codes
    
    # loop over types and try to get the grouping code
    for family_type_storage in family_storage_data.family_type_data_storage:
        if (isinstance(family_type_storage, FamilyTypeDataStorage)==False):
            raise TypeError("family_type_storage must be an instance of FamilyTypeDataStorage. Got instead: {}".format(type(family_type_storage)))
        
        grouping_code_parameter = family_type_storage.get_parameter_by_name(GROUPING_CODE_PARAMETER_NAME)

        # check if successful
        if (grouping_code_parameter is None):
            raise ValueError("Grouping code parameter '{}' not found in family type data storage.".format(GROUPING_CODE_PARAMETER_NAME))
        
        # remove any sub codes 
        group_code_cleaned = clean_code(grouping_code_parameter.value)
        if grouping_code_parameter.value not in unique_group_codes:
            unique_group_codes[grouping_code_parameter.value] = group_code_cleaned  # store the cleaned code as the value in the dictionary
           
    
    return unique_group_codes


def create_copy_directive_catalogue_file(name, category, source_file_path, target_directory, new_name ):
    """
    Create a copy directive for the catalogue file.

    :param name: The name of the catalogue file.
    :type name: str
    :param category: The category of the family.
    :type category: str
    :param source_file_path: The path to the source catalogue file.
    :type source_file_path: str
    :param target_directory: The directory where the catalogue file will be copied to.
    :type target_directory: str
    :param new_name: The new name for the catalogue file.
    :type new_name: str
    :return: A copy directive for the catalogue file.
    :rtype: :class:`.FamilyDirectiveCopy` or None if the source file does not exist.
    """

    # check if the source file path exists and is a file
    if (file_exist(source_file_path)):
        # create a copy directive for the catalogue file
        return FamilyDirectiveCopy(
            name=name, 
            category=category, 
            source_file_path=source_file_path, 
            target_directory=target_directory, 
            new_name=new_name,
        )
    return None

def create_copy_directives(family_storage_data, unique_group_codes, output_directory, code_to_descriptor_map, output):
    """
    Create copy directives for each unique group code in the family storage data.

    :param family_storage_data: The family storage data from which to create copy directives. (takes the family name, category and file path from this data)
    :type family_storage_data: :class:`.FamilyTypeDataStorageManager`
    :param unique_group_codes: A dictionary where key is the full code and value is the group code without any sub codes.
    :type unique_group_codes: dictionary {str:str}
    :param output_directory: The directory where the copied families will be saved.
    :type output_directory: str
    :param code_to_descriptor_map: A mapping of grouping codes to their descriptions.
    :type code_to_descriptor_map: dict[str, (str, str)]
    
    :return: A list of copy directives.
    :rtype: list[:class:`.FamilyDirectiveCopy`]
    """

    # set up a list containing all copy directives to be created
    copy_directives = []
    unique_new_file_name_list = []  # to keep track of unique new file names

    for full_code, group_code in unique_group_codes.items():
        output("...Processing full code: {} and group code: {}".format(full_code, group_code))
        
        # get the code to use to identify the family in the copy directive
        code_to_use, code_description = get_code_to_use(full_code, group_code, code_to_descriptor_map)

        # build family name from the group code and description
        fam_name_part = build_family_name_from_descriptor(code_description, code_to_use, output=output)

        output("...Family name part for group code {}: {} derived from: {}".format( code_to_use, fam_name_part, code_description))


        # set a default indicating that category does not exist in the mapper
        category_prefix = "WTF"
        # get the category prefix
        if family_storage_data.family_category in CATEGORY_FILE_NAME_PREFIX_MAPPER:
            # if the family category is in the mapper, use the prefix from the mapper
            category_prefix = CATEGORY_FILE_NAME_PREFIX_MAPPER[family_storage_data.family_category]

        # build new family name
        new_file_name_base ="{}_{}".format(category_prefix, fam_name_part)
        new_file_name_revit = "{}.rfa".format(new_file_name_base)
        new_family_name_catalogue = "{}.txt".format(new_file_name_base)

        # only add copy directive if the new file name is unique in the list of unique new file names
        # since we are combining multiple codes into one family there is a possibility that the new file name is not unique
        if new_file_name_revit in unique_new_file_name_list:
            continue  # skip if the new file name is not unique

        # add the new file name to the list of unique new file names
        unique_new_file_name_list.append(new_file_name_revit)

        output("Creating copy directive for group code: {} with new file name: {}".format(code_to_use, new_file_name_revit))
        # create a copy directive for each unique group code
        #name, category, source_file_path, target_directory, new_name
        copy_directive = FamilyDirectiveCopy(
            name = family_storage_data.family_name, 
            category = family_storage_data.family_category, 
            source_file_path = family_storage_data.family_file_path, 
            target_directory =  output_directory, 
            new_name = new_file_name_revit,
        )

        #output("...Created copy directive, new name: {}".format(copy_directive.new_name))
        
        # add to over all list
        copy_directives.append(copy_directive)

        # original catalogue file name
        original_catalogue_file_name = "{}.txt".format(family_storage_data.family_name[:-4])  # remove the '.rfa' from the family name

        # get the source file path
        original_catalogue_file_path = family_storage_data.family_file_path.replace(".rfa", ".txt")  # replace the '.rfa' with '.txt'

        # get catalogue file copy directive
        copy_directive_catalogue_file = create_copy_directive_catalogue_file( 
            original_catalogue_file_name,
            family_storage_data.family_category,  
            original_catalogue_file_path, 
            output_directory, 
            new_family_name_catalogue,
        )
        
        # check if the catalogue file copy directive is not None ( there is no catalogue file for this family )
        if copy_directive_catalogue_file is not None:
            #output("...Created copy directive, new name: {}".format(copy_directive_catalogue_file.new_name))
            # add to the copy directives list
            copy_directives.append(copy_directive_catalogue_file)

    # return the list of copy directives
    return copy_directives


def create_type_maintained_lists(family_storage_data, unique_group_codes, copy_directives, code_to_descriptor_map):

    """
    Create lists of family types to be maintained in the new family based on unique group codes.

    :param family_storage_data: The family storage data from which to create type maintained lists.
    :type family_storage_data: :class:`.FamilyTypeDataStorageManager`
    :param unique_group_codes: A dictionary where key is the full code and value is the group code without any sub codes.
    :type unique_group_codes: dictionary {str:str}
    :param copy_directives: A list of copy directives to be used for creating type maintained lists.
    :type copy_directives: list[:class:`.FamilyDirectiveCopy`]
    :param code_to_descriptor_map: A mapping of grouping codes to their descriptions.
    :type code_to_descriptor_map: dict[str, (str, str)]
    
    :return: A list of nested type maintained lists with two entries each: new family name, family type name to be maintained. ( a family with multiple times to be maintained will have multiple entries in the list )
    :rtype: list[ list[str] ]
    """
    
    #loop over unique group codes and
    # find associated copy directive
    # all types with matching group code

    type_keep_lists = []

    type_keep_list_unique = []

    for full_code, group_code in unique_group_codes.items():

        # get the code to use to identify the family in the copy directive
        code_to_use, code_description = get_code_to_use(full_code, group_code, code_to_descriptor_map)
        
        # get the code used on the family
        #group_code_descriptor_family = code_to_descriptor_map.get(group_code, None)
        # if group_code_descriptor_family is None:
        #    raise ValueError("Grouping code '{}' not found in code description mapping.".format(group_code))
        

        # amend the group code to the copy directive file name
        group_code_in_file_name = "{}.rfa".format(convert_to_file_name_code(code_to_use))

        # find the copy directive for this group code
        copy_directive = next((cd for cd in copy_directives if cd.new_name.endswith(group_code_in_file_name)), None)

        if copy_directive is not None:
            # loop over family types and add to type keep list
            for family_type_storage in family_storage_data.family_type_data_storage:

                # get the grouping code parameter from the family type storage
                grouping_code_parameter = family_type_storage.get_parameter_by_name(GROUPING_CODE_PARAMETER_NAME)
                
                # check if we have a grouping code parameter and if it matches the group code 
                if grouping_code_parameter :
                    if clean_code(grouping_code_parameter.value) == group_code or clean_code(grouping_code_parameter.value) == full_code:
                        key = "{}{}".format(copy_directive.new_name, family_type_storage.family_type_name)
                        # make sure we only add unique type keep lists
                        if key not in type_keep_list_unique:
                            type_keep_lists.append([copy_directive.new_name, family_type_storage.family_type_name])
                            type_keep_list_unique.append(key)  # add the key to the set of unique type keep lists

        else:
            print("Le impossibele: No copy directive found for group code: {}".format(group_code_in_file_name))
            for cp in copy_directives:
                print("Copy Directive: {}".format(cp.new_name))

    
    return type_keep_lists


def create_swap_directives(family_storage_data, unique_group_codes, copy_directives, code_to_descriptor_map, output=None):
    """
    Create swap directives for each unique group code in the family storage data.

    :param family_storage_data: The family storage data from which to create swap directives.
    :type family_storage_data: :class:`.FamilyTypeDataStorageManager`
    :param unique_group_codes: A dictionary of full code to group code mappings.(i.e ITSE-002.01: ITSE-002 or ITSE-001: ITSE-001)
    :type unique_group_codes: {str:str}
    :param copy_directives: A list of copy directives to be used for creating swap directives.
    :type copy_directives: list[:class:`.FamilyDirectiveCopy`]
    :param code_to_descriptor_map: A mapping of grouping codes to their descriptions.
    :type code_to_descriptor_map: dict[str, (str, str)]
    
    :return: A list of swap directives.
    :rtype: list[:class:`.FamilyDirectiveSwap`]
    """
    
    #loop over unique group codes and
    # find associated copy directive
    # build swap directive from old family name , new family name nad same family type name

    swap_directives = []
    swap_directives_unique = set()  # to keep track of unique swap directives

    for full_code, group_code in unique_group_codes.items():

        # get the code to use to identify the family in the copy directive
        code_to_use, code_description = get_code_to_use(full_code, group_code, code_to_descriptor_map)
        
        # amend the group code to the copy directive file name
        group_code_in_file_name = "{}.rfa".format(convert_to_file_name_code(code_to_use))
        # find the copy directive for this group code
        copy_directive = next((cd for cd in copy_directives if cd.new_name.endswith(group_code_in_file_name)), None)

        if copy_directive is not None:
            # loop over family types and create swap directives
            for family_type_storage in family_storage_data.family_type_data_storage:
                grouping_code_parameter = family_type_storage.get_parameter_by_name(GROUPING_CODE_PARAMETER_NAME)
                
                
                if grouping_code_parameter:
                    if clean_code(grouping_code_parameter.value) == group_code or clean_code(grouping_code_parameter.value) == full_code:
                        # check by family name, category and family type name if the swap directive is unique
                        key = "{}{}{}".format(family_storage_data.family_name, family_storage_data.family_category, family_type_storage.family_type_name)
                        if key not in swap_directives_unique:
                            swap_directive = FamilyDirectiveSwap(
                                name = family_storage_data.family_name,
                                category= family_storage_data.family_category,
                                source_type_name= family_type_storage.family_type_name,
                                target_family_name = copy_directive.new_name[:-4]   , # remove the '.rfa' !!
                                target_family_type_name = family_type_storage.family_type_name
                            )
                            swap_directives.append(swap_directive)
                            swap_directives_unique.add(key)  # add the key to the set of unique swap directives
                            if output is not None:
                                #output("Created swap directive for family'{}'.".format(key))
                                pass
                        else:
                            if output is not None:
                                #output("Swap directive for family type '{}' already exists, skipping...".format(key))
                                pass
                            #print("Swap directive for family type '{}' already exists, skipping...".format(family_type_storage.family_type_name))
        else:
            print("Le impossibele: No copy directive found for group code: {}".format(group_code_in_file_name))
            for cp in copy_directives:
                print("Copy Directive: {}".format(cp.new_name))


    return swap_directives


def create_directives(family_storage_data_list, output_directory, code_descriptor_path, output):
    """
    Create directives based on family storage data.

    :param family_storage_data: The family storage data from which to create directives.
    :type family_storage_data: a list of :class:`.FamilyTypeDataStorageManager`
    
    :return:
        Result class instance.

        - result.status: Directive creation status will be returned in result.status. False if an exception occurred, otherwise True.
        - result.message will be a log of conversion steps.
        - result.result will be a list containing the following items:
            - A list of :class:`.FamilyDirectiveCopy` directives for copying families.
            - A list of lists containing family type names to be maintained in the new family.
            - A list of :class:`.FamilyDirectiveSwap` directives for swapping family instances of types.
            - A list of family storage instances that had missing group codes in the code description mapping.

        On exception

        - Reload.status (bool) will be False
        - Reload.message will contain the exception message
        - Reload.result will be an empty list

    :rtype: :class:`.Result`
    """
    return_value = Result()
    
    try:
       # analyse each family:
       # set up a copy directive for each unique group code
       # set up a swap directive for each family type from old family to new family
       # create a list of types to keep per new family ( text file with same name as the new family name )

        overall_copy_directives = []
        overall_swap_directives = []
        overall_type_keep_lists = []
        families_with_missing_group_codes = []

        output("loading code description mapping from file: {}".format(code_descriptor_path))
        # load code  to descriptor mapper
        code_to_descriptor_map = load_group_code_description(code_descriptor_path)
        
        if code_to_descriptor_map is None:
            return_value.update_sep(
                False,
                "Failed to load code description mapping from file: {}".format(code_descriptor_path),
            )
            return return_value
        
        output("Loaded code to descriptor with {} entries from mapping from file.".format(len(code_to_descriptor_map)))

        # loop over all family storage instances and build directives
        for family_data_storage_instance in family_storage_data_list:

            return_value.append_message("Processing family storage data for family: {}".format(family_data_storage_instance.family_name))
            output("Processing family storage data for family: {}".format(family_data_storage_instance.family_name))
            unique_group_codes_in_family = {}
            
            # wrap into try catch since this may raise an exception if the family storage data is not valid or does not contain types
            try:
                # get the unique group codes from the family storage data
                # this function will raise an exception if the family storage data is not valid or does not contain types
                unique_group_codes_in_family = get_unique_group_codes_from_family(family_data_storage_instance)

                # if no codes found move on to the next family
                if len(unique_group_codes_in_family) == 0:
                    return_value.append_message(
                        "No unique group codes found in family storage data. {}".format(family_data_storage_instance.family_name),
                    )
                    families_with_missing_group_codes.append(family_data_storage_instance)
                    continue
            except Exception as e:
                return_value.append_message(
                    "Failed to get unique group codes from family storage data for family {}: {}. Skipping...".format(family_data_storage_instance.family_name, e),
                )
                output("Failed to get unique group codes from family storage data for family {}: {}. Skipping...".format(family_data_storage_instance.family_name, e))
                families_with_missing_group_codes.append(family_data_storage_instance)
                # skip to next family
                continue

            return_value.append_message("Found {} unique group codes in family storage data for family: {}".format(len(unique_group_codes_in_family), family_data_storage_instance.family_name))
            output("Found {} unique group codes in family storage data for family: {}".format(len(unique_group_codes_in_family), family_data_storage_instance.family_name))
            found_all_group_codes = True
            
            missing_group_codes = {}
            # check all group codes exist in descriptor mapper
            for code_full, group_code in unique_group_codes_in_family.items():
                if code_full not in code_to_descriptor_map and group_code not in code_to_descriptor_map:
                    return_value.append_message(
                        "Full code '{}' and group code: {} not found in code description mapping.".format(code_full,group_code),
                    )
                    output("......Full code '{}' and group code: {} not found in code description mapping.".format(group_code,group_code))
                    found_all_group_codes = False
                    missing_group_codes[code_full] = group_code # add the full code and group code to the list of codes
                    #break

            if not found_all_group_codes:
                # remove the missing codes from the unique group codes in family
                for code_full, group_code in missing_group_codes.items():
                    unique_group_codes_in_family.pop(code_full, None)  # remove the full code from the dictionary
                    return_value.append_message(
                        "......Full code '{}' and group code: {} removed due to missing mapping mapping.".format(code_full,group_code),
                    )
                    output("......Full code '{}' and group code: {} removed due to missing mapping mapping.".format(code_full,group_code))
                
            copy_directives = []
            try:
                # create directives for each unique group code
                copy_directives = create_copy_directives(family_data_storage_instance, unique_group_codes_in_family, output_directory, code_to_descriptor_map, output=output)
            except Exception as e:
                return_value.append_message(
                    "Failed to create copy directives for family {}: {}. Skipping...".format(family_data_storage_instance.family_name, e),
                )
                output("Failed to create copy directives for family {}: {}. Skipping...".format(family_data_storage_instance.family_name, e))
                # skip to next family
                continue
            
            return_value.append_message("Created {} copy directives for family: {}".format(len(copy_directives), family_data_storage_instance.family_name))
            output("Created {} copy directives for family: {}".format(len(copy_directives), family_data_storage_instance.family_name))
            
            # add directives to be returned
            overall_copy_directives = overall_copy_directives + copy_directives

            type_keep_lists = []

            try:
                # create lists of types to be maintained in the new family
                type_keep_lists = create_type_maintained_lists(
                    family_data_storage_instance, 
                    unique_group_codes_in_family, 
                    copy_directives, 
                    code_to_descriptor_map
                )
            except Exception as e:
                return_value.append_message(
                    "Failed to create type maintain lists for family {}: {}. Skipping...".format(family_data_storage_instance.family_name, e),
                )
                output("Failed to create type maintain lists for family {}: {}. Skipping...".format(family_data_storage_instance.family_name, e))
                # skip to next family
                continue

            return_value.append_message("Created {} type keep lists for family: {}".format(len(type_keep_lists), family_data_storage_instance.family_name))
            output("Created {} type keep lists for family: {}".format(len(type_keep_lists), family_data_storage_instance.family_name))

            overall_type_keep_lists = overall_type_keep_lists + type_keep_lists
            
            swap_directives = []

            try:
                # create swap directives
                swap_directives = create_swap_directives(
                    family_data_storage_instance, 
                    unique_group_codes_in_family, 
                    copy_directives, 
                    code_to_descriptor_map,
                    output=output,
                )
            except Exception as e:
                return_value.append_message(
                    "Failed to create swap directives for family {}: {}. Skipping...".format(family_data_storage_instance.family_name, e),
                )
                output("Failed to create swap directives for family {}: {}. Skipping...".format(family_data_storage_instance.family_name, e))
                # skip to next family
                continue

            return_value.append_message("Created {} swap directives for family: {}".format(len(swap_directives), family_data_storage_instance.family_name))
            output("Created {} swap directives for family: {}".format(len(swap_directives), family_data_storage_instance.family_name))

            overall_swap_directives = overall_swap_directives + swap_directives

        # return the overall copy directives, keep lists, swap directives, and list of instances with missing group codes
        return_value.result = []    
        return_value.result.append(overall_copy_directives)
        return_value.result.append(overall_type_keep_lists)
        return_value.result.append(overall_swap_directives)
        return_value.result.append(families_with_missing_group_codes)
        
    except Exception as e:
        output(e)
        return_value.update_sep(
            False,
            "Failed to create directives with exception: {}".format(e),
        )
    
    output("Directive creation completed successfully {}".format(return_value.status))
    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions to read copy  directives file(s) and return them to the caller
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

These helper function expect a text file in csv format with 5 columns:

- CURRENT_FAMILY_NAME = 0
- FAMILY_FILE_PATH = 1
- CATEGORY = 2
- NEW_FAMILY_NAME = 3
- NEW_DIRECTORY = 4

Note:

- There is no header row in the file, so all rows are treated as data rows.

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import os


from duHast.Utilities import files_csv as fileCSV, files_get as fileGet
from duHast.Utilities.files_io import copy_file, file_exist
from duHast.Utilities.Objects import result as res
from duHast.Revit.Family.Data.Objects.family_directive_copy import FamilyDirectiveCopy


def get_copy_directives(directory_path):
    """
    Retrieves file rename  directives from a given folder location.

    :param directory_path: Fully qualified folder path to folder containing directives.
    :type directory_path: str

    :return:
        Result class instance.

        - result.status. True if rename directives where found and loaded successfully, otherwise False.
        - result.message will contain number of directives found in format:'Found rename directives: ' + number
        - result.result list of directives

        On exception:

        - result.status (bool) will be False.
        - result.message will contain an exception message.
        - result.result will be empty

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    # check whether csv files matching file name filter exist in directory path
    rename_directive_files = fileGet.get_files_from_directory_walker_with_filters(
        directory_path,
        FamilyDirectiveCopy.COPY_DIRECTIVE_FILE_NAME_PREFIX,
        "",
        FamilyDirectiveCopy.COPY_DIRECTIVE_FILE_EXTENSION,
    )

    # check whether any files where found?
    if len(rename_directive_files) > 0:
        # attempt to re rename directives from files
        rename_directives =get_copy_directives(rename_directive_files)
        # check whether any rename directives where found in files
        if len(rename_directives) > 0:
            return_value.update_sep(
                True, "Found copy directives: {}".format(len(rename_directives))
            )
            # store rename directives in result object
            return_value.result = rename_directives
        else:
            return_value.update_sep(
                False, FamilyDirectiveCopy.EXCEPTION_EMPTY_COPY_DIRECTIVE_FILES
            )
    else:
        return_value.update_sep(
            False, FamilyDirectiveCopy.EXCEPTION_NO_COPY_DIRECTIVE_FILES
        )

    return return_value

def get_copy_directives(files):
    """
    Reads list of copy directives from files.

    :param filePath: List of fully qualified file path to copy and rename directives file.
    :type filePath: [str]
    :return: List of copy and rename directives.
    :rtype: [copy_and_rename_directive]
    """

    copy_and_rename_directives = []

    for file in files:
        rows_result = fileCSV.read_csv_file(file)
        # check whether file was read successfully
        if rows_result.status is False:
            return copy_and_rename_directives
        
        rows = rows_result.result
        
        # read rows in tuples ignoring the header row
        for i in range(0, len(rows)):
            data = None
            if len(rows[i]) == 5:
                data = FamilyDirectiveCopy(
                    name=rows[i][
                        FamilyDirectiveCopy. COPY_DIRECTIVE_LIST_INDEX_CURRENT_FAMILY_NAME
                    ],
                    source_file_path=rows[i][
                        FamilyDirectiveCopy.COPY_DIRECTIVE_INDEX_FAMILY_FILE_PATH
                    ],
                    category=rows[i][
                        FamilyDirectiveCopy.COPY_DIRECTIVE_INDEX_CATEGORY
                    ],
                    new_name=rows[i][
                        FamilyDirectiveCopy.COPY_DIRECTIVE_LIST_INDEX_NEW_FAMILY_NAME
                    ],
                    target_directory=rows[i][
                        FamilyDirectiveCopy.COPY_DIRECTIVE_LIST_INDEX_NEW_DIRECTORY
                    ],
                )
            else:
                continue

            copy_and_rename_directives.append(data)
    return copy_and_rename_directives


def execute_copy_directives(copy_directives):
    """
    Executes copy directives.

    :param copy_directives: List of copy directives to execute.
    :type copy_directives: [FamilyDirectiveCopy]
    :return: Result object with status and messages.
    :rtype: :class:`.Result`
    """

    # create result object
    return_value = res.Result()

    try:
        for directive in copy_directives:
            # execute copy directive
            if (file_exist(directive.source_file_path) is False):
                return_value.update_sep(
                    False,
                    "File does not exist: {}".format(directive.source_file_path),
                )
                continue


            target_file_path = None
            # attempt to copy the file
            if directive.target_directory is not None:
                target_file_path = os.path.join(
                    directive.target_directory, directive.new_name
                )
            else:
                return_value.update_sep(
                    False,
                    "No target directory path provided for copy directive: {}".format(directive.name))
                continue

            print("copy target file path: {}".format(target_file_path))
            # copy file to target directory
            copy_file_flag = copy_file(
                directive.source_file_path, target_file_path
            )
            
            if copy_file_flag is False:
                return_value.update_sep(
                    False,
                    "Failed to copy file from {} to {}".format(
                        directive.source_file_path, target_file_path
                    ),
                )
            else:
                return_value.append_message(
                    "Copied file from \n{} to \n{}".format(
                        directive.source_file_path, target_file_path
                    )
                )
           
        pass
    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to get family data with exception: {}".format(e),
        )
    
    return return_value


def write_copy_directives_to_file(copy_directives, file_path):
    """
    Writes copy directives to a specified file in CSV format.

    :param swap_directives: List of swap directives to write to file.
    :type swap_directives: list of FamilyDirectiveSwap
    :param file_path: Fully qualified file path where the directives will be written.
    :type file_path: str
    :return: Result object indicating success or failure of the write operation.
    :rtype: Result
    """
    
    return_value = res.Result()


    # loop over directives and convert them to a list of lists
    copy_directives_list = []
    try:
        for directive in copy_directives:

            copy_directives_list.append([
                directive.name,
                directive.source_file_path,
                directive.category,
                directive.new_name,
                directive.target_directory,
            ])

        # write the directives to the file
        return_value = fileCSV.write_report_data_as_csv(
            file_name=file_path,
            header = [],
            data=copy_directives_list,
            quoting=fileCSV.csv.QUOTE_MINIMAL,
        )

    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to write copy directives with exception: {}".format(e),
        )
    
    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Sheets
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from Autodesk.Revit.DB import BuiltInParameter


def get_sheet_number(sht):
    """
    Get the sheet number of a sheet
    :param sht: The sheet to get the number of
    :type sht: ViewSheet
    :return: The sheet number
    :rtype: str
    """

    try:
        return sht.get_Parameter(BuiltInParameter.SHEET_NUMBER).AsString()
    except:
        return None


def get_sheet_name(sht):
    """
    Get the sheet name of a sheet
    :param sht: The sheet to get the name of
    :type sht: ViewSheet
    :return: The sheet name
    :rtype: str
    """

    try:
        return sht.get_Parameter(BuiltInParameter.SHEET_NAME).AsString()
    except:
        return None


def get_sheet_num_name_comb(sht, num_first=True, separator=" - "):
    """
    Get the sheet number and name of a sheet and returns the the combination
    of the two. Typically helpful for logging/printing.

    :type sht: ViewSheet
    :param num_first: Whether the sheet number should be first in the string. Defaults to True
    :type num_first: bool
    :param splitter: The string to split the sheet number and name with. Defaults to ' - '
    :type splitter: str
    :return: The sheet number and name
    :rtype: str

    """

    num = get_sheet_number(sht)
    name = get_sheet_name(sht)

    if all([num, name]):
        if num_first:
            return num + separator + name
        else:
            return name + separator + num
    else:
        return None



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit curtain walls helper functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
import System
from System.Collections.Generic import List

from duHast.Revit.Common.common import get_ids_from_element_collector
from duHast.Revit.Common.parameter_get_utils import get_built_in_parameter_value

# import Autodesk
from Autodesk.Revit.DB import (
    BuiltInCategory,
    BuiltInParameter,
    ElementClassFilter,
    ElementMulticategoryFilter,
    FamilyInstance,
    FamilySymbol,
    FilteredElementCollector,
    Panel,
)

# -------------------------------------------- common variables --------------------
#: header used in reports
REPORT_CURTAINWALL_ELEMENTS_HEADER = [
    "HOSTFILE",
    "CURTAINWALL_ELEMENT_TYPEID",
    "ReplaceMeTYPENAME",
]
#: Built in family name for empty system panel
CURTAINWALL_PANEL_EMPTY_FAMILY_NAME = "Empty System Panel"
#: Built in family name for empty system panel
CURTAINWALL_PANEL_SYSTEM_FAMILY_NAME = "Empty System Panel"
#: Built in family name for V-shaped mullion
CURTAINWALL_MULLION_V_FAMILY_NAME = "V Corner Mullion"
#: Built in family name for circular mullion
CURTAINWALL_MULLION_CIRCULAR_FAMILY_NAME = "Circular Mullion"
#: Built in family name for quad corner mullion
CURTAINWALL_MULLION_QUAD_FAMILY_NAME = "Quad Corner Mullion"
#: Built in family name for L-shaped corner mullion
CURTAINWALL_MULLION_L_FAMILY_NAME = "L Corner Mullion"
#: Built in family name for rectangular mullion
CURTAINWALL_MULLION_RECT_FAMILY_NAME = "Rectangular Mullion"
#: Built in family name for trapezoid corner mullion
CURTAINWALL_MULLION_TRAPEZ_FAMILY_NAME = "Trapezoid Corner Mullion"

BUILTIN_TYPE_FAMILY_NAMES = [
    CURTAINWALL_PANEL_EMPTY_FAMILY_NAME,
    CURTAINWALL_PANEL_SYSTEM_FAMILY_NAME,
    CURTAINWALL_MULLION_V_FAMILY_NAME,
    CURTAINWALL_MULLION_CIRCULAR_FAMILY_NAME,
    CURTAINWALL_MULLION_QUAD_FAMILY_NAME,
    CURTAINWALL_MULLION_L_FAMILY_NAME,
    CURTAINWALL_MULLION_RECT_FAMILY_NAME,
    CURTAINWALL_MULLION_TRAPEZ_FAMILY_NAME,
]

#: category filter for all element filters by category
CURTAINWALL_ELEMENTS_CATEGORY_FILTER = List[BuiltInCategory](
    [
        BuiltInCategory.OST_CurtainWallPanels,
        BuiltInCategory.OST_CurtainWallMullions,
    ]
)

# --------------------------------------------- utility functions ------------------


def get_all_curtain_wall_element_types_by_category(doc):
    """
    Gets a filtered element collector of all curtain wall element types in the model:

    Filters by multiple categories.

    - curtain wall panels
    - curtain wall mullions
    - in place family symbols!


    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing curtain wall element types.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    multi_cat_filter = ElementMulticategoryFilter(CURTAINWALL_ELEMENTS_CATEGORY_FILTER)
    collector = (
        FilteredElementCollector(doc)
        .WherePasses(multi_cat_filter)
        .WhereElementIsElementType()
    )
    return collector


def build_curtain_wall_element_type_dictionary(collector, dic):
    """
    Returns the dictionary past in with keys and or values added retrieved from collector past in.

    Keys are built in curtain wall element type names.
    TODO: this code repeats across a number of modules. Use generic instead!

    :param collector: A filtered element collector containing curtain wall element types.
    :type collector: Autodesk.Revit.DB.FilteredElementCollector
    :param dic: A dictionary containing key: curtain wall element type name, value: list of ids belonging to that type.
    :type dic: dictionary (key str, value list of Autodesk.Revit.DB.ElementId)

    :return: A dictionary containing key: built in curtain wall element type  name, value: list of ids belonging to that type.
    :rtype: dictionary (key str, value list of Autodesk.Revit.DB.ElementId)
    """

    for c in collector:
        if c.FamilyName in dic:
            if c.Id not in dic[c.FamilyName]:
                dic[c.FamilyName].append(c.Id)
        else:
            dic[c.FamilyName] = [c.Id]
    return dic


def sort_curtain_wall_element_types_by_family_name(doc):
    """
    Returns a dictionary containing all curtain wall element types in the model.

    Key values are as per BUILTIN_TYPE_FAMILY_NAMES.
    TODO: This code repeats across a number of modules. Use generic instead!

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A dictionary containing key: curtain wall element type family name, value: list of ids.
    :rtype: dic { str: [Autodesk.Revit.DB.ElementId]}
    """

    # get all CurtainWallElement types including in place wall families
    wts_two = get_all_curtain_wall_element_types_by_category(doc)
    used_wts = {}
    used_wts = build_curtain_wall_element_type_dictionary(wts_two, used_wts)
    return used_wts


# -------------------------------- none in place or loadable Curtain Wall Element types -------------------------------------------------------


def get_curtain_wall_element_instances_by_category(doc):
    """
    Gets all CurtainWallElement elements instances placed in model.

    Includes:

    - curtain wall panels
    - curtain wall mullions

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing curtain wall element types.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    multi_cat_filter = ElementMulticategoryFilter(CURTAINWALL_ELEMENTS_CATEGORY_FILTER)
    return (
        FilteredElementCollector(doc)
        .WherePasses(multi_cat_filter)
        .WhereElementIsNotElementType()
    )


def get_all_curtain_wall_element_type_ids_by_category(doc):
    """
    Gets all Curtain Wall Element element type ids available in model.

    Includes:

    - curtain wall panels
    - curtain wall mullions

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of element ids representing curtain wall element types.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = []
    col_cat = get_all_curtain_wall_element_types_by_category(doc)
    ids = get_ids_from_element_collector(col_cat)
    return ids


def get_all_curtain_wall_element_types_by_category_excl_in_place(doc):
    """
    Gets all Curtain Wall Element element type available in model. Excludes in place family symbols.

    Includes:

    - curtain wall panels
    - curtain wall mullions

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of curtain wall element types.
    :rtype: list of Autodesk.Revit.DB.ElementType
    """

    collector = get_all_curtain_wall_element_types_by_category(doc)
    elements = []
    for c in collector:
        if c.GetType() != FamilySymbol:
            elements.append(c)
    return elements


def get_all_curtain_wall_element_type_ids_by_category_excl_symbols(doc):
    """
    Gets all Curtain Wall Element element type ids available in model. Excludes in place family symbols.

    Includes:

    - curtain wall panels
    - curtain wall mullions

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of element ids representing curtain wall element types.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    collector = get_all_curtain_wall_element_types_by_category(doc)
    ids = []
    for c in collector:
        if c.GetType() != FamilySymbol:
            ids.append(c.Id)
    return ids


# -------------------------------- loadable Curtain Wall Element types -------------------------------------------------------


def get_all_curtain_wall_non_shared_symbol_ids_by_category(doc):
    """
    Gets a list of all loadable, non shared, family symbols (types) in the model of categories:

    - curtain wall panels
    - curtain wall mullions

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of element ids representing curtain wall family symbols.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = []
    multi_cat_filter = ElementMulticategoryFilter(CURTAINWALL_ELEMENTS_CATEGORY_FILTER)
    collector = (
        FilteredElementCollector(doc)
        .WherePasses(multi_cat_filter)
        .WhereElementIsElementType()
    )
    for c in collector:
        if c.GetType() == FamilySymbol:
            fam = c.Family
            p_value = get_built_in_parameter_value(fam, BuiltInParameter.FAMILY_SHARED)
            if p_value != None and p_value == "No" and c.Id not in ids:
                ids.append(c.Id)
    return ids


def get_curtain_wall_panels(doc, curtain_wall_instance):
    """
    Returns a list of panel instances that are hosted in a curtain wall, excluding any family instances.

    :param doc: The current Revit model document.
    :type doc: Document
    :param curtain_wall_instance: The curtain wall instance for which to retrieve the panel instances.
    :type curtain_wall_instance: FamilyInstance
    :return: A list of panel instances hosted in the curtain wall.
    :rtype: list[Panel]
    """

    panels = []
    # get the panels from the curtain walls
    filter = ElementClassFilter(FamilyInstance)
    dependent_family_instance_ids = curtain_wall_instance.GetDependentElements(filter)
    # loop over dependent elements and only return any panel instances
    for id in dependent_family_instance_ids:
        dependent_element = doc.GetElement(id)
        if type(dependent_element) == Panel:
            panels.append(dependent_element)
    return panels



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit grids appearance modifier functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Note: Any grid appearance modification in a view will throw an exception if the grid is not actually visible in the view.

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from Autodesk.Revit.DB import DatumEnds, DatumExtentType, Transaction


from duHast.Utilities.Objects import result as res
from duHast.Revit.Common import transaction as rTran


def change_grids_2D(doc, grids, view):
    """
    Changes all grids in view to 2D

    Note: Any grid past in, which is not visible in the view, will throw an exception when attempting to set to 2D.

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document
    :param grids: List of grids to be changed to 2D.
    :type grids: [Autodesk.Revit.DB.Grid]
    :param view: The view in which to change the grids
    :type view: Autodesk.Revit.DB.View
    :return:
        Result class instance.
        - result.status. True if all grids where set to 2D, otherwise False.
        - result.message will contain the name(s) of the grid(s) changed to 2D
        - result.result empty list
        On exception:
        - result.status (bool) will be False.
        - result.message will contain generic exception message including the grid name.
        - result.result will be empty
    :rtype: :class:`.Result`
    """

    return_value = res.Result()

    # needs to run in a transaction
    def action():
        action_return_value = res.Result()
        grid_counter = 0
        for g in grids:
            grid_counter = grid_counter + 1
            try:
                g.SetDatumExtentType(DatumEnds.End1, view, DatumExtentType.ViewSpecific)
                g.SetDatumExtentType(DatumEnds.End0, view, DatumExtentType.ViewSpecific)
                action_return_value.update_sep(
                    True, "Changed grid {} to 2D.".format(g.Name)
                )
            except Exception as e:
                action_return_value.update_sep(
                    False,
                    "Failed to change grid {} to 2D with exception: {}".format(
                        g.Name, e
                    ),
                )
        if grid_counter == 0:
            action_return_value.update_sep(
                True, "No grids visible in view {}".format(view.Name)
            )
        return action_return_value

    transaction = Transaction(doc, "Grids to 2D")
    return_value = rTran.in_transaction(transaction, action)
    return return_value


def show_bubble_end(doc, grid, view, end_identifier, show_bubble):
    """
    Toggles grid bubble visibility on the specified end for a given grid.

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document

    :param grid: The grid of which a bubble's visibility is to be toggled.
    :type grid: Autodesk.Revit.DB.Grid

    :param view: The view in which a grid bubble's visibility is to be toggled.
    :type view: Autodesk.Revit.DB.View

    :param end_identifier: An enumerated type representing ends of a datum plane.
    :type end_identifier: Autodesk.Revit.DB.DatumEnds

    :param show_bubble: True if the bubble will be switched on, False if it will be switched off.
    :type show_bubble: bool

    :return: Result class instance.

        - `result.status` (bool): True if the grid bubble(s) visibility was set successfully, otherwise False.
        - `result.message` (str): Contains the name(s) of the grid(s) where bubble visibility was set.
        - `result.result` (list): Empty list.

    On exception:

        - `result.status` (bool) will be False.
        - `result.message` (str) will contain a generic exception message including the grid name.
        - `result.result` (list) will be empty.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()

    # needs to run in a transaction
    def action():
        action_return_value = res.Result()
        try:
            if show_bubble:
                grid.ShowBubbleInView(end_identifier, view)
                action_return_value.update_sep(
                    True,
                    "Set grid {} bubble to visible at end: {}".format(
                        grid.Name, end_identifier
                    ),
                )
            else:
                grid.HideBubbleInView(end_identifier, view)
                action_return_value.update_sep(
                    True,
                    "Set grid {} bubble to invisible at end: {}".format(
                        grid.Name, end_identifier
                    ),
                )
        except Exception as e:
            action_return_value.update_sep(
                False,
                "Failed to change grid {} bubble visibility at end {} with exception: {}".format(
                    grid.Name, end_identifier, e
                ),
            )
        return action_return_value

    transaction = Transaction(doc, "Toggle Bubble. {}".format((show_bubble)))
    return_value = rTran.in_transaction(transaction, action)
    return return_value


def hide_both_bubbles(doc, grids, view):
    """
    Hides both bubbles of grids in the given view.

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document

    :param grid: The grid for which bubble visibility is to be toggled.
    :type grid: Autodesk.Revit.DB.Grid

    :param view: The view in which grid bubble visibility is to be toggled.
    :type view: Autodesk.Revit.DB.View

    :return: Result class instance.

        - `result.status` (bool): True if all grid bubble(s) visibility was switched off successfully, otherwise False.
        - `result.message` (str): Contains the name(s) of the grid(s) where bubble visibility was set.
        - `result.result` (list): Empty list.

    On exception:

        - `result.status` (bool) will be False.
        - `result.message` (str) will contain a generic exception message including the grid name.
        - `result.result` (list) will be empty.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    for g in grids:
        return_value.update(show_bubble_end(doc, g, view, DatumEnds.End1, False))
        return_value.update(show_bubble_end(doc, g, view, DatumEnds.End0, False))

    return return_value


def show_bubble_zero_end(doc, grids, view):
    """
    Turns on grid bubbles at the zero end in the specified view.

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document

    :param grid: The grids for which bubble visibility is to be toggled.
    :type grid: List[Autodesk.Revit.DB.Grid]

    :param view: The view in which grid bubble visibility is to be toggled.
    :type view: Autodesk.Revit.DB.View

    :return: Result class instance.

        - `result.status` (bool): True if all grid bubble(s) visibility at the zero end was set to visible successfully, otherwise False.
        - `result.message` (str): Contains the name(s) of the grid(s) where bubble visibility was set.
        - `result.result` (list): Empty list.

    On exception:

        - `result.status` (bool) will be False.
        - `result.message` (str) will contain a generic exception message including the grid name.
        - `result.result` (list) will be empty.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    for g in grids:
        return_value.update(show_bubble_end(doc, g, view, DatumEnds.End0, True))

    return return_value


def show_bubble_one_end(doc, grids, view):
    """
    Turns on grid bubbles at one end in the specified view.

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document

    :param grid: The grids for which bubble visibility is to be toggled.
    :type grid: List[Autodesk.Revit.DB.Grid]

    :param view: The view in which grid bubble visibility is to be toggled.
    :type view: Autodesk.Revit.DB.View

    :return: Result class instance.

        - `result.status` (bool): True if all grid bubble(s) visibility at one end was set to visible successfully, otherwise False.
        - `result.message` (str): Contains the name(s) of the grid(s) where bubble visibility was set.
        - `result.result` (list): Empty list.

    On exception:

        - `result.status` (bool) will be False.
        - `result.message` (str) will contain a generic exception message including the grid name.
        - `result.result` (list) will be empty.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    for g in grids:
        return_value.update(show_bubble_end(doc, g, view, DatumEnds.End1, True))

    return return_value


def toggle_bubble_end(doc, grid, view, end_identifier):
    """
    Toggles grid bubble visibility on a specified end for a given grid in given views.

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document

    :param grid: The grid of which bubble visibility is to be toggled.
    :type grid: Autodesk.Revit.DB.Grid

    :param view: The view in which grid bubble visibility is to be toggled.
    :type view: Autodesk.Revit.DB.View

    :param end_identifier: The end of the grid to be modified.
    :type end_identifier: Autodesk.Revit.DB.DatumEnds

    :return: Result class instance.

        - `result.status` (bool): True if the grid bubble(s) visibility was changed successfully, otherwise False.
        - `result.message` (str): Contains the name of the grid where bubble visibility was changed.
        - `result.result` (list): Empty list.

    On exception:

        - `result.status` (bool) will be False.
        - `result.message` (str) will contain a generic exception message including the grid name.
        - `result.result` (list) will be empty.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()

    def action():
        try:
            action_return_value = res.Result()
            endBubbleOne = grid.IsBubbleVisibleInView(end_identifier, view)
            if endBubbleOne == False:
                grid.ShowBubbleInView(end_identifier, view)
                action_return_value.update_sep(
                    True,
                    "Set grid {} bubble to visible at end: {}.".format(
                        grid.Name, end_identifier
                    ),
                )
            else:
                grid.HideBubbleInView(end_identifier, view)
                action_return_value.update_sep(
                    True,
                    "Set grid {} bubble to not visible at end: {}.".format(
                        grid.Name, end_identifier
                    ),
                )
        except Exception as e:
            action_return_value.update_sep(
                False,
                "Failed to change grid {} bubble visibility at end: {} with exception: {}".format(
                    grid.Name, end_identifier, e
                ),
            )
        return action_return_value

    transaction = Transaction(doc, "Toggle Bubble.")
    return_value = rTran.in_transaction(transaction, action)
    return return_value


def toggle_bubble_one_end(doc, grids, view):
    """
    Toggles grid bubble visibility on one end for given grids.

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document

    :param grids: The grids of which bubble visibility at one end is to be toggled.
    :type grids: List[Autodesk.Revit.DB.Grid]

    :param view: The view in which grid bubble visibility is to be toggled.
    :type view: Autodesk.Revit.DB.View

    :return: Result class instance.

        - `result.status` (bool): True if the bubble visibility at one end was changed successfully for all grids, otherwise False.
        - `result.message` (str): Contains the name(s) of the grid(s) where bubble visibility was changed.
        - `result.result` (list): Empty list.

    On exception:

        - `result.status` (bool) will be False.
        - `result.message` (str) will contain a generic exception message including the grid name.
        - `result.result` (list) will be empty.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    for g in grids:
        return_value.update(toggle_bubble_end(doc, g, view, DatumEnds.End1))
    return return_value


def toggle_bubble_zero_end(doc, grids, view):
    """
    Toggles grid bubble visibility on zero end for given grids.

    :param doc: The current model document.
    :type doc: Autodesk.Revit.DB.Document

    :param grids: The grids of which bubble visibility at zero end is to be toggled.
    :type grids: List[Autodesk.Revit.DB.Grid]

    :param view: The view in which grid bubble visibility is to be toggled.
    :type view: Autodesk.Revit.DB.View

    :return: Result class instance.

        - `result.status` (bool): True if the bubble visibility at zero end was changed successfully for all grids, otherwise False.
        - `result.message` (str): Contains the name(s) of the grid(s) where bubble visibility was changed.
        - `result.result` (list): Empty list.

    On exception:

        - `result.status` (bool) will be False.
        - `result.message` (str) will contain a generic exception message including the grid name.
        - `result.result` (list) will be empty.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    for g in grids:
        return_value.update(toggle_bubble_end(doc, g, view, DatumEnds.End0))
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A number of utility functions which create views usually found on a Room Layout Sheet
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Views created for a room are:

- plan
- elevations
- 3D view
- schedule



"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects.result import Result
from duHast.Data.Objects.Collectors.Properties.data_view_port_type_names import DataViewPortTypeNames
from duHast.Revit.Views.plan_view_create import create_view_plan
from duHast.Revit.Views.elevation_view_create import create_elevation_views
from duHast.Revit.Rooms.Objects.RoomSpatialObject import RoomSpatialObj
from duHast.Utilities.date_stamps import get_date_stamp, TIME_STAMP_HHMMSEC_UNDERSCORE


from duHast.Revit.Common.transaction import in_transaction_with_failure_handling
from duHast.Revit.Common.Objects.FailureHandlingConfiguration import (
    FailureHandlingConfig,
)
from duHast.Revit.Common.parameter_set_utils import set_builtin_parameter_without_transaction_wrapper_by_name


from Autodesk.Revit.DB import  BuiltInParameter,Transaction

def create_plan_view_from_room (doc, room, view_plan_type):
    return_value = Result()

    # in order to create the view plan I need:
    # - level ( same as room)
    # - view plan type
    # - phase of the view ( same as the room created)
    # - bounding box: same as the rooms
    # - view name : room number + room name

    try:

        # test if room is placed in model ...just in case
        if room.Location is None:
            raise ValueError ("Room is not placed in model.")
        
        # get the room information
        room_spatial = RoomSpatialObj(rvt_doc=doc, room=room)

        # create a plan view
        create_view_status = create_view_plan(
            doc=doc, 
            level=room_spatial.level, 
            view_type=view_plan_type, 
            view_name="{} {}".format(room_spatial.number_name_comb, get_date_stamp(TIME_STAMP_HHMMSEC_UNDERSCORE)),
            phasing = room_spatial.phase_object, 
            view_crop=room_spatial.bbox,
            view_crop_visible=False,
        )

        return_value.update(create_view_status)

    except Exception as e:
        return_value.update_sep(False, "An exception occured: {}".format(e))

    return return_value


def create_elevation_views_in_room (doc, room, view_plan, view_elevation_type, scale):

    """
    Creates elevations in a room

    based on: # https://forums.autodesk.com/t5/revit-api-forum/automate-room-elevations-using-revit-api/td-p/3504490
    
    :return: _description_
    :rtype: _type_
    """
    
    return_value = Result()

    try:
        # test if room is placed in model ...just in case
        if room.Location is None:
            raise ValueError ("Room is not placed in model.")
        
        # get the room information
        room_spatial = RoomSpatialObj(rvt_doc=doc, room=room)


        # to speed things up I will manage the transaction here

        def action():
            action_return_value = Result()
            try:
                
                # create an elevation marker first with a single elevation
                result_marker = create_elevation_views(
                    doc=doc,
                    view_type=view_elevation_type,
                    view_marker_location=room_spatial.bbox_centre,
                    scale=scale,
                    view_index_list=[3],
                    view_names=[],
                    view_plan=view_plan,
                    phasing=room_spatial.phase_object,
                    view_crop=None,
                    view_crop_visible=False,
                    transaction_manager=None
                )

                action_return_value.update(result_marker)

                if result_marker.status == False:
                    raise ValueError("Creating elevation marker failed.")
                
                # get the elevation marker
                elevation_marker = result_marker.result[0]

                # get the single elevation view
                view_three = result_marker.result[1]

                   

                # rotate the marker so it faced the X axis ( from min_x, min_y to max_x, min_y)
                # I eventually worked out that I needed to create one of the elevations (index 3), rotate the marker, create elevation indexes 0, 1, & 2, and then delete 3, and then recreate 3.
                # What? Why? By using this approach the extents of the elevations are set correctly, i.e, extend to surrounding walls.

                # TODO: rotate the elevation if required
                # You can check the orientation of any view by getting View.ViewDirection. To reorient the ElevationMarker and all of the views it hosts, call ElementTransformUtils.RotateElement with the ElevationMarker as an argument.




                # create all other markers but 3
                view_zero = elevation_marker.CreateElevation(doc, view_plan.Id, 0)
                view_one = elevation_marker.CreateElevation(doc, view_plan.Id, 1)
                view_two = elevation_marker.CreateElevation(doc, view_plan.Id, 2)

                # delete marker 3
                doc.Delete(view_three.Id)

                # create elevation for 3
                view_three = elevation_marker.CreateElevation(doc, view_plan.Id, 3)

                # rename views to match room number and name
                view_zero.Name = "{} {}".format(room_spatial.number_name_comb, "Elevation 0")
                action_return_value.append_message("View Zero Name: {}".format(view_zero.Name))
                view_one.Name = "{} {}".format(room_spatial.number_name_comb, "Elevation 1")
                action_return_value.append_message("View One Name: {}".format(view_one.Name))
                view_two.Name = "{} {}".format(room_spatial.number_name_comb, "Elevation 2")
                action_return_value.append_message("View Two Name: {}".format(view_two.Name))
                view_three.Name = "{} {}".format(room_spatial.number_name_comb, "Elevation 3")
                action_return_value.append_message("View Three Name: {}".format(view_three.Name))


                # set the phasing of the views (this is most likely drive by the view plan...but just in case)
                result_phase = set_builtin_parameter_without_transaction_wrapper_by_name(
                    element=view_zero, 
                    parameter_definition=BuiltInParameter.VIEW_PHASE, 
                    parameter_value=room_spatial.phase_object.Id
                )
                action_return_value.append_message("Setting phase for view zero: {} with status {}".format(result_phase.message, result_phase.status))
                
                result_phase = set_builtin_parameter_without_transaction_wrapper_by_name(
                    element=view_one,
                    parameter_definition=BuiltInParameter.VIEW_PHASE,
                    parameter_value=room_spatial.phase_object.Id
                )
                action_return_value.append_message("Setting phase for view one: {} with status {}".format(result_phase.message, result_phase.status))

                result_phase = set_builtin_parameter_without_transaction_wrapper_by_name(
                    element=view_two,
                    parameter_definition=BuiltInParameter.VIEW_PHASE,
                    parameter_value=room_spatial.phase_object.Id
                )
                action_return_value.append_message("Setting phase for view two: {} with status {}".format(result_phase.message, result_phase.status))

                result_phase = set_builtin_parameter_without_transaction_wrapper_by_name(
                    element=view_three,
                    parameter_definition=BuiltInParameter.VIEW_PHASE,
                    parameter_value=room_spatial.phase_object.Id
                )

                action_return_value.append_message("Setting phase for view three: {} with status {}".format(result_phase.message, result_phase.status))

                # do I need to check 
                # - elevation depth - only want to look to the wall not beyond it
                # - elevation line location ( 1m out from wall max )

                # return the created views (and reset the return value)
                action_return_value.result = [view_zero, view_one, view_two, view_three]
               
            except Exception as e:
                action_return_value.update_sep(False, "An exception occured: {}".format(e))
            
            return action_return_value

        # execute the action
        # define failure handling for the transaction ( roll back on any warnings or errors )
        failure_handling_settings = FailureHandlingConfig(
            roll_back_on_warning=True,
            print_warnings=False,
            roll_back_on_error=True,
            print_errors=False,
        )
       
        # set up the transaction
        trans = Transaction(doc, "creating view elevation(s) {}")

        # execute the transaction with failure handling
        result_transaction = in_transaction_with_failure_handling(
            transaction=trans,
            action=action,
            failure_config=failure_handling_settings,
        )

        # update the return value with the result of the transaction
        return_value.update(result_transaction)

    except Exception as e:
        return_value.update_sep(False, "An exception occured: {}".format(e))

    return return_value

def create_3D_view_of_room(doc, room):
    return_value = Result()

    try:
        pass
    except Exception as e:
        return_value.update_sep(False, "An exception occured: {}".format(e))

    return return_value

def create_schedule_for_room_content(doc, room):
    return_value = Result()

    try:
        pass
    except Exception as e:
        return_value.update_sep(False, "An exception occured: {}".format(e))

    return return_value

def create_views(doc, room, view_plan_type, view_elevation_type):

    return_value = Result()

    views_created = {}

    try:
        return_value.append_message("Creating plan view...")
        # create plan view
        create_plan_view_status = create_plan_view_from_room(doc=doc, room=room, view_plan_type=view_plan_type)
        return_value.update(create_plan_view_status)
        # store the created view
        if(create_plan_view_status.status):
            views_created[DataViewPortTypeNames.FLOOR_PLAN] = create_plan_view_status.result

        # create elevations
        return_value.append_message("Creating elevation views...")
        create_elevation_views_status = create_elevation_views_in_room(
            doc=doc, 
            room=room, 
            view_plan=views_created[DataViewPortTypeNames.FLOOR_PLAN][0],
            view_elevation_type=view_elevation_type,
            scale=50)
        
        return_value.update(create_elevation_views_status)
        # store the created views
        if(create_elevation_views_status.status):
            views_created[DataViewPortTypeNames.ELEVATION] = create_elevation_views_status.result

        # create 3D view
        create_3D_view_status = create_3D_view_of_room(doc=doc, room=room)
        return_value.update(create_3D_view_status)
        # store the created view
        if(create_3D_view_status.status):
            views_created[DataViewPortTypeNames.THREE_D] = create_3D_view_status.result

        # create schedule
        create_schedule_status = create_schedule_for_room_content(doc=doc, room=room)
        return_value.update(create_schedule_status)
        # store the created view
        if(create_schedule_status.status):
            views_created[DataViewPortTypeNames.SCHEDULE] = create_schedule_status.result

        # overwrite any objects the return value object will ahve at this point
        # keep to the standard format of a list
        return_value.result = [views_created]


    except Exception as e:
        return_value.update_sep(False, "Creating views of room failed with exception: {}".format(e))

    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Family elements
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Family.Data.Objects.family_base_data_processor_defaults import (
    NESTING_SEPARATOR,
)

from Autodesk.Revit.DB import Element, FilteredElementCollector, Family


def get_name_to_family_dict(rvt_doc):
    """
    Create a dictionary of family name and the Family element
    :param rvt_doc: Revit document
    :type rvt_doc: Autodesk.Revit.DB.Document
    :return: Dictionary of family name and Family element
    :rtype: dict
    """

    # Get all the families in the model
    all_families = FilteredElementCollector(rvt_doc).OfClass(Family).ToElements()
    # create a dictionary of family name and family object
    family_dict = {fam.Name: fam for fam in all_families}
    return family_dict


def get_name_and_category_to_family_dict(rvt_doc):
    """
    Create a dictionary of family name and category concatenated and the Family element.
    
    Note:
        - The category is concatenated with the family name using the NESTING_SEPARATOR.
    
    This is useful when there are multiple families with the same name but different categories.

    :param rvt_doc: Revit document
    :type rvt_doc: Autodesk.Revit.DB.Document
    :return: Dictionary of family name and category and Family element
    :rtype: dict
    """

    # Get all the families in the model
    all_families = FilteredElementCollector(rvt_doc).OfClass(Family).ToElements()
    
    # create a dictionary of family name and family object
    family_dict = {"{}{}{}".format(fam.Name, NESTING_SEPARATOR, fam.FamilyCategory.Name): fam for fam in all_families}
    
    return family_dict


def get_symbol_names_of_family(family):
    """
    Get all the symbol names of a family

    :param family: Family element
    :type family: Autodesk.Revit.DB.Family
    :return: List of symbol names of the family
    :rtype: list
    """

    symbol_ids = [sym for sym in family.GetFamilySymbolIds()]
    doc = family.Document
    symbol_names = [
        Element.Name.GetValue(doc.GetElement(sym_id)) for sym_id in symbol_ids
    ]
    return symbol_names



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A module with helper function to read xml family type data into storage objects.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Separate module for ease of testing and maintainability.
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


import clr
import datetime
import re

clr.AddReference("System.Xml")
from System.Xml import XmlNamespaceManager, XmlDocument

from duHast.Revit.Family.Data.Objects.family_type_parameter_data_storage import (
    FamilyTypeParameterDataStorage,
)
from duHast.Revit.Family.Data.Objects.family_type_data_storage import (
    FamilyTypeDataStorage,
)
from duHast.Revit.Family.Data.Objects.family_type_data_storage_manager import (
    FamilyTypeDataStorageManager,
)
from duHast.Revit.Family.Data.Objects.family_base_data_processor_defaults import (
    NESTING_SEPARATOR,
)
from duHast.Utilities.utility import encode_ascii

from duHast.Utilities.string_operations import (
    remove_currency_sign,
    replace_new_lines,
    remove_trailing_characters_from_number_string,
)

# Constants
# place holder for unitless parameters
UNITLESS = "unitless"
NAME_UNKNOWN = "unknown name"
TYPE_UNKNOWN = "unknown type" # type of parameter is unknown (options are custom, shared, system)
VALUE_UNKNOWN = "unknown value"
PARAMETER_STORAGE_TYPE_UNKNOWN = "unknown storage type"
FAMILY_TYPE_NAME_UNKNOWN = "unknown family type name"


# XML node attribute names translation of parameter properties
# the parameter name
PARAMETER_NAME = "Parameter Name"
# the parameter type (options are custom, shared, system)
PARAMETER_INTERNAL_TYPE = "Parameter Internal Type"
# the storage type of a parameter (string, yes /no,etc)
PARAMETER_STORAGE_TYPE = "Parameter Storage Type"
# the units of a parameter
PARAMETER_UNITS = "Parameter Units"

# XML node name mapper of properties for parameters
# the translated properties of a parameter and the xml attribute name
PARAMETER_NODE_PROPERTIES = {
    PARAMETER_NAME: "name",
    PARAMETER_INTERNAL_TYPE:"type",
    PARAMETER_STORAGE_TYPE: "typeOfParameter",
    PARAMETER_UNITS: "units",
}

# XML node name containing the family type name
CHILD_NODE_NAME_CONTAINING_FAMILY_TYPE_NAME = "title"

# XML node name containing the family types
CHILD_NODE_NAME_CONTAINING_FAMILY_TYPES = "//A:part"
# XML node name containing the family data
CHILD_NODE_NAME_CONTAINING_FAMILY_DATA = "//A:family"
# XML node name containing the latest update information (date and time)
CHILD_NODE_NAME_CONTAINING_UPDATE_DATA = "//atom:updated"
# XML node name containing the family category data
CHILD_NODE_NAME_CONTAINING_FAMILY_CATEGORY = "//atom:category"

#XML namespace for the atom feed
NAME_SPACES = {
    "atom":"http://www.w3.org/2005/Atom",
    "A": "urn:schemas-autodesk-com:partatom"
}


def get_name_space_manager(doc_xml):
    """
    Get the XML namespace manager for the XML document. Also adds the required namespaces.

    :param doc_xml: The XML document.
    :type doc_xml: XmlDocument

    :return: An XML namespace manager.
    :rtype: XmlNamespaceManager
    """

    name_space_manager = XmlNamespaceManager(doc_xml.NameTable)

    for name_space in NAME_SPACES:
        name_space_manager.AddNamespace(name_space, NAME_SPACES[name_space])

    return name_space_manager


def get_parameter(xml_node, family_name, root_category_path, family_path, family_type_name = FAMILY_TYPE_NAME_UNKNOWN, default_value=VALUE_UNKNOWN):
    """
    Get a parameter from the XML node.

    :param xml_node: The XML node.
    :type xml_node: XmlNode
    :param family_name: The name of the family.
    :type family_name: str
    :param root_category_path: The root category path.
    :type root_category_path: str
    :param family_path: The path of the family file.
    :type family_path: str
    :param family_type_name: The name of the family type.
    :type family_type_name: str
    :param default_value: The default value to use if the parameter value is not set.
    :type default_value: str

    :return: A family type parameter data storage object.
    :rtype: :class:`.FamilyTypeParameterDataStorage`
    """

    try:
        # attempt to read out values
        name = NAME_UNKNOWN
        try:
            name = xml_node.Name
        except Exception as e:
            name = "{}".format(name, e)

        type = TYPE_UNKNOWN
        try:
            type = xml_node.Attributes[PARAMETER_NODE_PROPERTIES[PARAMETER_INTERNAL_TYPE]].Value
        except Exception as e:
            type = "{}".format(type, e)

        type_of_parameter = PARAMETER_STORAGE_TYPE_UNKNOWN
        try:
            type_of_parameter = xml_node.Attributes[PARAMETER_NODE_PROPERTIES[PARAMETER_STORAGE_TYPE]].Value
        except Exception as e:
            type_of_parameter = "{}".format(type_of_parameter, e)

        # there are parameters without units (i.e. text parameters)
        units = UNITLESS
        try:
            units = xml_node.Attributes[PARAMETER_NODE_PROPERTIES[PARAMETER_UNITS]].Value
        except Exception as e:
            pass

        # set the parameter value to the default value
        p_value = default_value

        # attempt to read out values if required
        if p_value == VALUE_UNKNOWN:
        
            try:
                # replace any new row characters with space and remove trailing spaces
                p_value = replace_new_lines(xml_node.InnerText)
            except Exception as e:
                pass

            # check if the value is a number and contains thousands separators
            if (
                type_of_parameter
                in FamilyTypeParameterDataStorage.unit_type_compare_values_as_floats
            ):
                # remove any thousands separators
                p_value = p_value.replace(",", "")
                # remove any currency signs
                p_value = remove_currency_sign(p_value)
                # remove any trailing units
                p_value = remove_trailing_characters_from_number_string(p_value)
        else:
            # go with the default value
            pass

        # family_name can either be just the family name or the family root path
        # check which one it is:
        root_name_path = family_name
        family_name_checked = family_name

        # split the file name at nesting separator
        family_name_split = family_name.split(NESTING_SEPARATOR)
        
        # check if the family name contains a nesting separator
        if len(family_name_split) > 0:
            # yes, so the fam name path is the last part
            family_name_checked = family_name_split[-1]

        # Create a parameter object 
        # make sure all values are encoded to ascii to avoid 
        # issues with special characters when writing to file
        parameter = FamilyTypeParameterDataStorage(
            root_name_path = encode_ascii(root_name_path),
            root_category_path = encode_ascii(root_category_path),
            family_name = encode_ascii(family_name_checked),
            family_type_name=encode_ascii(family_type_name),
            family_file_path=encode_ascii(family_path),
            name=encode_ascii(name),
            type=encode_ascii(type),
            type_of_parameter=encode_ascii(type_of_parameter),
            units=units, # not encoding because some units are not ascii
            value=encode_ascii(p_value),
        )

        return parameter
    except Exception:
        return None



def get_unique_parameters_from_family_xml(doc_xml, family_name, root_category_path, family_path):
    """
    Get all unique parameters from the family type XML document by inspecting every part node representing a family type.
    
    Note:
    The part atom export will only contain parameters for a family type if they have a value set.

    :param doc_xml: The XML document.
    :type doc_xml: XmlDocument

    :return: A list of family type parameter data storage objects.
    :rtype: [:class:`.FamilyTypeParameterDataStorage`]
    """

    if isinstance(doc_xml, XmlDocument) is False:
        raise TypeError(
            "doc_xml must be an instance of XmlDocument. Got: {}".format(doc_xml)
        )

    parameters = []

    # Add an XML namespace manager
    name_space_manager = get_name_space_manager(doc_xml)

    # Get the family parameters
    for part_node in doc_xml.SelectNodes(CHILD_NODE_NAME_CONTAINING_FAMILY_TYPES, name_space_manager):
        for child_node in part_node.ChildNodes:
            # check for family type nodes
            if child_node.Name != CHILD_NODE_NAME_CONTAINING_FAMILY_TYPE_NAME:
                # get the parameter
                parameter = get_parameter(
                    xml_node=child_node, 
                    family_name=family_name,
                    root_category_path= root_category_path, 
                    family_path=family_path,
                    family_type_name="",
                    default_value=""
                )
                
                # Add parameter to list if not in there already
                if parameter is not None and parameter not in parameters:
                    parameters.append(parameter)

    return parameters
                   
def add_missing_empty_parameters(all_parameters_in_atom_export, parameters_in_type, root_name_path, root_category_path, family_name, family_path, family_type_name):
    """
    Add missing empty parameters to the list of parameters for a family type.

    :param all_parameters_in_atom_export: The list of all parameters in the atom export.
    :type all_parameters_in_atom_export: [:class:`.FamilyTypeParameterDataStorage`]
    :param parameters_in_type: The list of parameters for the family type.
    :type parameters_in_type: [:class:`.FamilyTypeParameterDataStorage`]
    :param root_name_path: The root name path.
    :type root_name_path: str
    :param root_category_path: The root category path.
    :type root_category_path: str
    :param family_name: The name of the family.
    :type family_name: str
    :param family_path: The path of the family file.
    :type family_path: str
    :param family_type_name: The name of the family type.
    :type family_type_name: str

    :return: The list of parameters for the family type.
    :rtype: [:class:`.FamilyTypeParameterDataStorage`]
    """
    # loop over all parameters in the atom export and check whether they are in the parameters list for this type
    for parameter_in_export in all_parameters_in_atom_export:
        # check if the parameter is in the list of parameters for this type
        parameter_found = False
        for parameter_in_type in parameters_in_type:
            if parameter_in_export.name == parameter_in_type.name:
                parameter_found = True
                break
        
        # if the parameter is not in the list of parameters for this type add it with an empty value
        if parameter_found == False:
            parameter_missing = FamilyTypeParameterDataStorage(
                root_name_path = root_name_path,
                root_category_path = root_category_path,
                family_name = family_name,
                family_type_name=family_type_name,
                family_file_path=family_path,
                name=parameter_in_export.name,
                type=parameter_in_export.type,
                type_of_parameter=parameter_in_export.type_of_parameter,
                units=parameter_in_export.units,
                value="",
            )
            parameters_in_type.append(parameter_missing)
            # reset the parameter found flag
            parameter_found = False
    
    return parameters_in_type


def read_xml_into_storage(doc_xml, family_name, family_path, root_category_path = "None"):
    """
    Read the XML data into the storage object.

    Note all names and values will be encoded to ascii.

    :param doc_xml: The XML document.
    :type doc_xml: XmlDocument
    :param family_name: The name of the family.
    :type family_name: str
    :param family_path: The path of the family file.
    :type family_path: str
    :param root_category_path: The root category path.
    :type root_category_path: str

    :return: A family type data storage manager object.
    :rtype: :class:`.FamilyTypeDataStorageManager`
    """

    if isinstance(doc_xml, XmlDocument) is False:
        raise TypeError(
            "doc_xml must be an instance of XmlDocument. Got: {}".format(doc_xml)
        )

    type_data_storage_manager = FamilyTypeDataStorageManager()
    # Add an XML namespace manager
    name_space_manager = get_name_space_manager(doc_xml)

    # Select the family node
    family_node = doc_xml.SelectSingleNode(CHILD_NODE_NAME_CONTAINING_FAMILY_DATA , name_space_manager)

    # check if category root path is not set, if so ignore use the one from the xml
    if root_category_path == "None":
        # Get the category nodes ( there will be more than one)
        for cat_node in doc_xml.SelectNodes(CHILD_NODE_NAME_CONTAINING_FAMILY_CATEGORY, name_space_manager):
            dummy_term = ""
            dummy_scheme = ""

            for child_node in cat_node.ChildNodes:
                if child_node.Name == "term":
                    dummy_term = child_node.InnerText
                if child_node.Name == "scheme":
                    dummy_scheme = child_node.InnerText
            # check if this is the category name
            if dummy_scheme == "adsk:revit:grouping":
                root_category_path = dummy_term

    # get the date and time of the last update
    last_updated_date = None
    last_updated_time = None

    # Select the <updated> node directly under the <entry> node
    updated_node = doc_xml.SelectSingleNode(CHILD_NODE_NAME_CONTAINING_UPDATE_DATA, name_space_manager)
    if updated_node is not None:
        last_updated_datetime = updated_node.InnerText

        # Convert the date-time string to a date and time
        try:
            dt = datetime.datetime.strptime(last_updated_datetime, "%Y-%m-%dT%H:%M:%SZ")
            last_updated_date = dt.date().isoformat()  # Convert to date-only string
            last_updated_time = dt.time().isoformat()  # Convert to time-only string
        except ValueError as e:
            print("Error parsing date-time: {}".format(e))
    else:
        print("updated_node not found")

    # it turns out that an empty parameter value means that parameter is not listed in the part atom export file for that type
    # it might be listed for another type though where it has a value...

    # need to loop over types first to ensure that all parameters are read in
    all_parameters_in_atom_export = get_unique_parameters_from_family_xml(doc_xml, family_name, root_category_path, family_path)

    # Get the family parameters
    for part_node in family_node.SelectNodes(CHILD_NODE_NAME_CONTAINING_FAMILY_TYPES, name_space_manager):

        # Get the family type name
        family_type_name = None
        for child_node in part_node.ChildNodes:
            if child_node.Name == CHILD_NODE_NAME_CONTAINING_FAMILY_TYPE_NAME:
                family_type_name = child_node.InnerText
                break

        # If we got a type name, add the parameters, their values and units, parameter type and type of parameter
        if family_type_name:

            parameters = []
            for child_node in part_node.ChildNodes:
                if child_node.Name != CHILD_NODE_NAME_CONTAINING_FAMILY_TYPE_NAME:

                    # get the parameter
                    parameter = get_parameter(
                        xml_node=child_node, 
                        family_name=family_name,
                        root_category_path=root_category_path, 
                        family_path=family_path,
                        family_type_name=family_type_name,
                        default_value=VALUE_UNKNOWN, # enforce parameter value to be read in
                    )

                    # Add parameter to list if valid
                    if parameter is not None:
                        # Add type to family
                        parameters.append(parameter)
            
            # check if there are parameters in the atom export that are not in the parameters list for this type since they have no value for this family type set
            parameters = add_missing_empty_parameters(
                all_parameters_in_atom_export=all_parameters_in_atom_export, 
                parameters_in_type=parameters,
                root_name_path=encode_ascii(family_name),
                root_category_path=encode_ascii(root_category_path),
                family_name=encode_ascii(family_name),
                family_path=encode_ascii(family_path),
                family_type_name=encode_ascii(family_type_name),
            )

            # Set up a family type data storage object
            # make sure all values are encoded to ascii to avoid 
            # issues with special characters when writing to file
            fam_type = FamilyTypeDataStorage(
                root_name_path=encode_ascii(family_name),
                root_category_path=encode_ascii(root_category_path),
                family_name=encode_ascii(family_name),
                family_file_path=encode_ascii(family_path),
                family_type_name=encode_ascii(family_type_name),
                parameters=parameters,
                last_updated_date=last_updated_date,
                last_updated_time=last_updated_time,
            )

            # Add the family type to the storage manager for this family
            type_data_storage_manager.add_family_type_data_storage(fam_type)

    # set the family name, category, file path if they are not set yet
    if type_data_storage_manager.family_name is None:
        type_data_storage_manager.family_name = family_name
    if type_data_storage_manager.family_category is None:
        type_data_storage_manager.family_category = root_category_path
    if type_data_storage_manager.family_file_path is None:
        type_data_storage_manager.family_file_path = family_path

    return type_data_storage_manager



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to:
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

- date stamps (with varies formatting options)

"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


#: default file stamp date format using underscores as delimiter: 21_03_01
import datetime

from duHast.Utilities.files_io import get_file_name_without_ext


FILE_DATE_STAMP_YY_MM_DD = "%y_%m_%d"


#: file stamp date format using spaces as delimiter: 21 03 01
FILE_DATE_STAMP_YYMMDD_SPACE = "%y %m %d"


#: file stamp date format using spaces as delimiter: 2021 03 01
FILE_DATE_STAMP_YYYYMMDD_SPACE = "%Y %m %d"


#: file stamp date format using underscores as delimiter: 2021_03_01
FILE_DATE_STAMP_YYYY_MM_DD = "%Y_%m_%d"


#: file stamp date time format using underscores as delimiter: 2021_03_01_18_59_59
FILE_DATE_STAMP_YYYY_MM_DD_HH_MM_SEC = "%Y_%m_%d_%H_%M_%S"

#: file stamp date time format using underscores as delimiter: 2021_03_01_18_59_59_123
FILE_DATE_STAMP_YYYY_MM_DD_HH_MM_SEC_MSEC = "%Y_%m_%d_%H_%M_%S_%f"


#: time stamp using colons: 18:59:59
TIME_STAMP_HHMMSEC_COLON = "%H:%M:%S"

#: time stamp using colons: 18_59_59
TIME_STAMP_HHMMSEC_UNDERSCORE = "%H_%M_%S"


def get_file_date_stamp(format=FILE_DATE_STAMP_YY_MM_DD):
    """
    Returns a date stamp formatted suitable for a file name.
    :param format: The date stamp format, defaults to FILE_DATE_STAMP_YY_MM_DD
    :type format: str, optional
    :return: datetime.now() string formatted using supplied format string
    :rtype: str
    """

    d = datetime.datetime.now()
    return d.strftime(format)


#: folder date format: no delimiter 210301
FOLDER_DATE_STAMP_YYMMDD = "%y%m%d"


#: folder date format: no delimiter 20210301
FOLDER_DATE_STAMP_YYYYMMDD = "%Y%m%d"


#: folder date format: no delimiter 2021
FOLDER_DATE_STAMP_YYYY = "%Y"

def date_time_stamp_builder(format):
    """
    Returns a formatted date stamp.

    Note: milliseconds, if required, are capped to 3 digits.

    :param format: The date stamp format
    :type format: str
    :return: datetime.now() string formatted using supplied format string
    :rtype: str
    """

    if isinstance(format, str):
        # check if the format is a valid date format
        try:
            d = datetime.datetime.now()
            formatted_time = d.strftime(format)
            
            # Check if the format includes microseconds and trim to milliseconds if needed
            if "%f" in format:
                formatted_time = formatted_time[:-3]

            return formatted_time
        except ValueError:
            raise ValueError("Invalid date format string provided.")
    else:
        raise TypeError("Format must be a string, got {}".format(type(format))) 
    

def get_folder_date_stamp(format=FOLDER_DATE_STAMP_YYYYMMDD):
    """
    Returns a date stamp formatted suitable for a folder name.

    Note: milliseconds, if required, are capped to 3 digits.

    :param format: The date stamp format, defaults to FOLDER_DATE_STAMP_YYYYMMDD
    :type format: str, optional
    :return: datetime.now() string formatted using supplied format string
    :rtype: str
    """

    return  date_time_stamp_builder(format)



def get_date_stamp(format):
    """
    Returns a date stamp formatted using past in format string.

    Note: milliseconds, if required, are capped to 3 digits.

    :param format: The date stamp format
    :type format: str
    :return: datetime.now() string formatted using supplied format string
    :rtype: str
    """

    return date_time_stamp_builder(format)


def get_date_stamped_file_name(file_path, file_extension=".txt", file_suffix=""):
    """
    Returns a time stamped output file name based on the past in file name and file extension.
    :param file_path: Fully qualified file path to file
    :type file_path: str
    :param file_extension: File extension needs to include '.', defaults to '.txt'
    :type file_extension: str, optional
    :param file_suffix: File suffix will be appended after the name but before the file extension, defaults to ''
    :type file_suffix: str, optional
    :return: File name.
    :rtype: str
    """

    # get date prefix for file name
    file_prefix = get_file_date_stamp()
    name = get_file_name_without_ext(file_path)
    return file_prefix + "_" + name + file_suffix + file_extension



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions to load or reload families 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

# class used for stats reporting
from duHast.Utilities.Objects import result as res

from duHast.Revit.Family.family_utils import load_family
from duHast.Revit.Common.delete import delete_by_element_ids


def reload_family(doc, family, family_file_path):
    """
    Reloads a single family into a Revit document.

    Will reload family provided in in path. By default the type parameter values in the project file will be overwritten
    with type parameter values in family.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param family: The family to be reloaded
    :type family: Autodesk.Revit.DB.family
    :param family_file_path: The fully qualified file path of the family to be re-loaded.
    :type family_file_path: str
    :raise: None

    :return:
        Result class instance.

        - Reload status (bool) returned in result.status.
        - Reload message: contains the reload log messages.
        - Return family reference stored in result.result property on successful reload only (single entry in list)

        On exception

        - Reload.status (bool) will be False
        - Reload.message will contain the exception message

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    try:

        # get all symbols attached to this family by name
        prior_load_symbol_ids = family.GetFamilySymbolIds()

        # load the family
        load_result = load_family(doc=doc, family_file_path=family_file_path)

        # preserve reload log messages
        return_value.append_message(load_result.message)

        # do I need to get that family from the reload or can I get it from the project again??
        # check if a family was returned from the reload
        # indicating I need to check for new types
        if(len(load_result.result)>0):
            
            # get the family returned
            fam_loaded = load_result.result[0]
            # get all its symbol ids 
            after_load_symbol_ids = fam_loaded.GetFamilySymbolIds()
            # find all new symbols introduced during the reload
            new_symbol_ids =  [item for item in after_load_symbol_ids if item not in prior_load_symbol_ids]
            
            # if any new symbols, delete them
            if len(new_symbol_ids) > 0:
                # delete the new symbols
                result_delete = delete_by_element_ids(
                    doc,
                    new_symbol_ids,
                    "Delete new family types",
                    "Family types",
                )
                # preserve the delete outcome
                return_value.append_message(result_delete.message)
            else:
                # make note nothing needed to be deleted
                return_value.append_message("The reload did not add any new types to the project.")
            
            # return the family in the result object
            return_value.result.append(fam_loaded)

    except Exception as e:
        return_value.update_sep(False, "Failed to load families with exception: {}".format(e))
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family Base data processing utility module.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Processes family_data_family objects using multi threading ( testing has shown results in multi threading are not as expected, so single core processing is used for now)

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import threading
import os
import queue

from duHast.Utilities.Objects import result as res
from duHast.Revit.Family.Data.family_report_reader import read_data_into_families


# def process_data(family_base_data_report_file_path, do_this):
#     """
#     Process family data using multi threading

#     :param family_base_data_report_file_path: str: path to directory containing family report files
#     :param do_this: function: function taking two args: list of family_data_family and result list

#     :return:
#         Result class instance.

#         - result.status. True if processing completed without an exception occurring.
#         - result.message will contain the summary messages of the process including time stamps.
#         - result.result [:class:`.FamilyDataFamily`]

#         On exception:

#         - result.status (bool) will be False.
#         - result.message will contain generic exception message.
#         - result.result will be empty

#     :rtype: :class:`.Result`
#     """

#     # do_this: function taking two args: list of family_data_family and result list
    

#     return_value = res.Result()
#     # results will be stored in here:
#     processing_result = []

#     try:
#         # read families into data
#         read_result = read_data_into_families(family_base_data_report_file_path)

#         return_value.update(read_result)
#         return_value.append_message(
#             "Number of family instances: {} read.".format(
#                 len(read_result.result),
#             )
#         )

#         # check if something went wrong
#         if not read_result.status:
#             raise ValueError(read_result.message)
#         elif len(read_result.result) == 0:
#             raise ValueError(
#                 "No family data found in file: {}".format(
#                     family_base_data_report_file_path
#                 )
#             )

#         # read results
#         return_value.result.append(read_result.result)

#         # set up some multithreading
#         core_count = int(os.environ["NUMBER_OF_PROCESSORS"])
#         return_value.append_message("cores: {}".format(core_count))

#         if core_count > 2:
#             # leave some room for other processes
#             core_count = core_count - 1
#             return_value.append_message("using threads: {}".format(core_count))
#             # attempt int division
#             chunk_size = len(read_result.result) // core_count
#             return_value.append_message("...chunk size: {}".format(chunk_size))
#             threads = []
#             start_value = 0
#             end_value = start_value + chunk_size
#             # set up threads
#             for i in range(core_count):
#                 if start_value + chunk_size <= len(read_result.result) - 1:
#                     end_value = start_value + chunk_size
#                 else:
#                     end_value = len(read_result.result)
#                 return_value.append_message(
#                     "......assigning chunk {} : {} of {}".format(
#                         start_value, end_value, len(read_result.result)
#                     )
#                 )
#                 t = threading.Thread(
#                     target=do_this,
#                     args=(
#                         read_result.result[start_value:end_value],
#                         processing_result,
#                     ),
#                 )
#                 threads.append(t)

#                 # increase the start value for the next processing chunk by 1 to avoid duplicate processing
#                 start_value = end_value + 1

#             # start up threads
#             for t in threads:
#                 t.start()
#             # wait for results
#             for t in threads:
#                 t.join()
#         else:
#             # no threading
#             processing_result = do_this(read_result.result, processing_result)
#     except Exception as e:
#         return_value.update_sep(
#             False, "Failed to read and process families with exception: {}".format(e)
#         )

#     return_value.result = processing_result
#     return return_value

def process_data(family_base_data_report_file_path, do_this):
    return_value = res.Result()
    processing_results = []

    try:
        read_result = read_data_into_families(family_base_data_report_file_path)
        return_value.update(read_result)
        return_value.append_message(
            "Number of family instances: {} read.".format(len(read_result.result))
        )

        if not read_result.status:
            raise ValueError(read_result.message)
        elif len(read_result.result) == 0:
            raise ValueError(
                "No family data found in file: {}".format(family_base_data_report_file_path)
            )

        return_value.result.append(read_result.result)
        core_count = int(os.environ.get("NUMBER_OF_PROCESSORS", 1))
        return_value.append_message("cores: {}".format(core_count))

        # if core_count > 2:
        #     core_count -= 1
        #     return_value.append_message("using threads: {}".format(core_count))
        #     chunk_size = len(read_result.result) // core_count
        #     return_value.append_message("...chunk size: {}".format(chunk_size))
        #     threads = []
        #     start_value = 0

        #     for i in range(core_count):
        #         end_value = min(start_value + chunk_size, len(read_result.result))
        #         return_value.append_message(
        #             "......assigning chunk {} : {} of {}".format(start_value, end_value, len(read_result.result))
        #         )
        #         thread_result = []
        #         t = threading.Thread(
        #             target=do_this,
        #             args=(read_result.result[start_value:end_value], thread_result)
        #         )
        #         threads.append((t, thread_result))
        #         t.start()
        #         start_value = end_value + 1

        #     for t, thread_result in threads:
        #         t.join()
        #         processing_results.extend(thread_result)
        # else:
        
        # go single core for now only...multi thread processing is not working
        result = do_this(read_result.result, processing_results)
        return_value.result = result
    except Exception as e:
        return_value.update_sep(False, "Failed to read and process families with exception: {}".format(e))

    return_value.result = processing_results
    return return_value